import math
import random
from typing import Any, Iterable

# from ...text.utils import novelity_score
from .abstract_language_model import LanguageModel


class BeamSampling:
    def __init__(
        self,
        model: LanguageModel,
        beam_width: int = 3,
        start_of_sequence_token=["["],  # ["<s>"],
        end_of_sequence_token=["]"],  # ["</s>"],
        max_length=33,
        scoring_function=lambda seq, log_prob: 10.0 + log_prob / len(seq),
        return_log_probability=True,
    ) -> None:
        self.model = model
        self.start_of_sequence_token = start_of_sequence_token
        self.end_of_sequence_token = end_of_sequence_token
        self.beam_width = beam_width
        self.max_length = max_length
        self.scoring_function = scoring_function
        self.return_log_probability = return_log_probability

    def __call__(self, context: Iterable = None) -> Iterable:
        return self.complete(context=context)

    def concatenate(self, context, token):
        if isinstance(context, str):
            joint = context + token
        elif isinstance(context, list):
            joint = context + [token]
        elif isinstance(context, tuple):
            joint = context + (token,)
        # numpy / torch

        return joint

    def complete(self, context: Iterable = None) -> Iterable:
        if context is None:
            context = self.start_of_sequence_token

        branches = [(context, 0.0)]

        for _ in range(self.max_length):
            n_branches = round(self.beam_width + random.random() * 0.8 - 0.4)
            # Expand
            branches_ = []
            for context_, score_ in branches:
                if (
                    # ended
                    context_[-1] == self.end_of_sequence_token
                    or len(context_) >= self.max_length
                ):
                    branches_.append((context_, score_))
                else:
                    for _ in range(n_branches):
                        next_token, prob = self.model.next(context_)
                        if next_token == self.model.unknown_token:
                            score = score_
                            next_context = context_
                        else:
                            next_context = self.concatenate(context_, next_token)
                            score = score_ + math.log(prob)
                        branches_.append((next_context, score))

            # Prune
            branches_ = sorted(
                branches_,
                key=lambda item: self.scoring_function(*item),
                reverse=True,
            )[:n_branches]

            if branches == branches_:
                branches = branches_
                break
            else:
                branches = branches_

        weights = [self.scoring_function(seq, log_prob) for seq, log_prob in branches]

        # softmax
        weights = [math.exp(w) for w in weights]
        weights = [w / sum(weights) for w in weights]

        selected_completion, score = random.choices(
            branches,
            weights=weights,
            k=1,
        )[0]

        if not self.return_log_probability:
            score = math.exp(score)

        return selected_completion, score



from abc import ABC, abstractmethod, abstractstaticmethod
from typing import Any, Iterable, List, Set, Tuple, Union


class TextLanguage(ABC):
    @abstractstaticmethod
    def name() -> str:
        pass

    @abstractstaticmethod
    def word_regex() -> str:
        pass

    @abstractstaticmethod
    def allowed_characters() -> Set[str]:
        pass

    @abstractmethod
    def preprocess(self, text: str) -> str:
        pass

    @abstractmethod
    def tokenize(self, text: str) -> List[str]:
        pass

    @abstractmethod
    def sentence_tokenize(self, text: str) -> List[str]:
        pass

    @abstractmethod
    def detokenize(self, tokens: Iterable[str]) -> str:
        pass

    @abstractmethod
    def tag(self, tokens: Union[str, Iterable[str]]) -> List[Tuple[str, Any]]:
        pass

    @abstractmethod
    def get_tags(self, tokens: Union[str, Iterable[str]]) -> List[Any]:
        pass

    @abstractmethod
    def get_word_senses(self, tokens: Union[str, Iterable[str]]) -> List[List[str]]:
        pass

    # embed/similar
    # all_tags


from abc import ABC, abstractmethod
from typing import Any, Iterable, Tuple


class LanguageModel(ABC):
    def __init__(self, unknown_token="<unk>", name=None) -> None:
        super().__init__()
        self.unknown_token = unknown_token
        self.name = name

    def __call__(self, context: Iterable) -> Tuple[Iterable, float]:
        """Samples the next token from the model based on the given context.

        Args:
            context (Iterable): A piece of sequence like the training examples.

        Returns:
            Tuple[Any, float]: The next token and its associated probability. Token has the same type as the context iterable.
        """

        return self.next(context)

    @abstractmethod
    def next(self, context: Iterable) -> Tuple[Any, float]:
        pass

    @abstractmethod
    def next_all(self, context: Iterable) -> Tuple[Iterable[Any], Iterable[float]]:
        pass

    def __str__(self) -> str:
        return (
            f'name="{self.name}", ' if self.name else ""
        ) + f'unk_tok="{self.unknown_token}"'

    def __repr__(self) -> str:
        return str(self.__class__).split()[1][:-1] + "\n\n" + self.__str__()



"""Module that contains Text Languages Processors as classes to clean up, tokenize and tag texts"""

from sign_language_translator.languages.text.text_language import TextLanguage
from sign_language_translator.languages.text.urdu import Urdu
from sign_language_translator.text.tagger import Tags

__all__ = [
    "TextLanguage",
    "Urdu",
    "Tags",
]



"""
This module provides classes for mapping rules used in a mapping system.

The mapping system allows for mapping tokens to specific objects based on predefined rules.
The module includes abstract base classes for mapping rules and several concrete implementations.

Classes:
- MappingRule: Abstract base class for mapping rules.
- LambdaMappingRule: Mapping rule defined by lambda functions.
- DirectMappingRule: Mapping rule for supported words.
- CharacterByCharacterMappingRule: Mapping rule for character-by-character mapping.

Each mapping rule class defines the behavior of the rule,
including applicability checks and actions to be taken when the rule is applied.
"""

from abc import ABC, abstractmethod
from typing import Any, Callable, Dict, List, Set


class MappingRule(ABC):
    """
    Abstract base class for mapping rules.
    """

    @property
    @abstractmethod
    def priority(self) -> int:
        """
        Priority of the mapping rule.

        Returns:
            int: The priority of the mapping rule.
        """

    @abstractmethod
    def is_applicable(
        self,
        token: str,
        tag: Any | None = None,
        context: Any | None = None,
    ) -> bool:
        """
        Check if the mapping rule is applicable for the given token, tag, and context.

        Args:
            token (str): The token to be checked.
            tag (Any, optional): The tag associated with the token. Defaults to None.
            context (Any, optional): The context in which the token appears. Defaults to None.

        Returns:
            bool: True if the mapping rule is applicable, False otherwise.
        """

    @abstractmethod
    def apply(self, token: str) -> Any:
        """
        Apply the mapping rule to the given token. (i.e. map the given token to something.)

        Args:
            token (str): The token to apply the mapping rule to.

        Returns:
            The result of applying the mapping rule.
        """


class LambdaMappingRule(MappingRule):
    """Mapping rule based on lambda functions.

    Args:
        is_applicable_function (Callable[[str, Any, Any], bool]):
            Function to check if the rule is applicable to the given token, tag & context.
        apply_function (Callable[[str], Any]): Function to apply the rule to the token.
        priority (int): Priority of the rule.
    """

    def __init__(
        self,
        is_applicable_function: Callable[[str, Any, Any], bool],
        apply_function: Callable[[str], Any],
        priority: int,
    ) -> None:
        super().__init__()
        self.is_applicable_function = is_applicable_function
        self.apply_function = apply_function
        self._priority = priority

    def is_applicable(self, token, tag=None, context=None) -> bool:
        return self.is_applicable_function(token, tag, context)

    def apply(self, token) -> Any:
        return self.apply_function(token)

    @property
    def priority(self):
        return self._priority


class DirectMappingRule(MappingRule):
    """Mapping rule that directly maps keys to values.

    Args:
        token_to_object (Dict[str, Any]): Dictionary mapping tokens to some objects.
        priority (int): Priority of the rule.
    """

    def __init__(self, token_to_object: Dict[str, Any], priority: int) -> None:
        super().__init__()
        self.token_to_object = token_to_object
        self._priority = priority

    def is_applicable(self, token, tag=None, context=None) -> bool:
        return token in self.token_to_object

    def apply(self, token: str) -> Any:
        return self.token_to_object[token]

    @property
    def priority(self):
        return self._priority


class CharacterByCharacterMappingRule(MappingRule):
    """Mapping rule which maps a token character-by-character.

    Args:
        token_to_object (Dict[str, Any]): Dictionary mapping tokens to some objects.
        allowed_tags (Set[Any]): Set of allowed tags for the rule to be applicable.
        priority (int): Priority of the rule.
    """

    def __init__(
        self,
        token_to_object: Dict[str, Any],
        allowed_tags: Set[Any],
        priority: int,
    ) -> None:
        super().__init__()
        self.token_to_object = token_to_object
        self.allowed_tags = allowed_tags
        self._priority = priority

    def is_applicable(self, token, tag=None, context=None) -> bool:
        return tag in self.allowed_tags and all(
            char in self.token_to_object for char in token
        )

    def apply(self, token) -> List[Any]:
        """Apply the mapping rule to the given token.

        Args:
            token (str): The token to apply the mapping rule to.

        Returns:
            List[Any]: list of results of applying the mapping rule to each character in the token.
        """

        return [self.token_to_object[char] for char in token]

    @property
    def priority(self):
        return self._priority



import os
import random
from typing import Any, Iterable, List, Tuple

from .abstract_language_model import LanguageModel


class Mixer(LanguageModel):
    def __init__(
        self,
        models: List[LanguageModel],
        selection_probabilities: List[float] = None,
        unknown_token="<unk>",
        name=None,
    ) -> None:

        if selection_probabilities is None:
            selection_probabilities = [1] * len(models)

        assert len(models) == len(selection_probabilities)
        assert all(isinstance(p, (float, int)) for p in selection_probabilities)

        super().__init__(unknown_token=unknown_token)
        self.models = models
        total = sum(selection_probabilities)
        self.selection_probabilities = [w / total for w in selection_probabilities]
        self.name = name

    def next(self, context: Iterable) -> Tuple[Iterable, float]:
        # if strategy = "merge"
        #       use finetune weighted_merge for all possible next[_all] token vectors then choose

        # if strategy == "choose"
        tried_models = set()
        while len(tried_models) < len(self.models):
            model_idx = random.choices(
                range(len(self.models)),
                weights=self.selection_probabilities,
                k=1,
            )[0]
            if model_idx in tried_models:
                continue
            else:
                tried_models.add(model_idx)

            token, prob = self.models[model_idx].next(context)
            if token != self.models[model_idx].unknown_token:
                return token, prob

        return self.unknown_token, 0.0

    def next_all(self, context: Iterable) -> Tuple[Iterable[Any], Iterable[float]]:
        return [], []

    def load(self, model_path: str) -> None:
        # folder
        pass

    def save(self, model_path: str) -> None:
        # folder
        pass

    def __str__(self) -> str:
        model_strs = []
        for i, model in enumerate(self.models):
            branch = "├── " if i < len(self.models) - 1 else "└── "
            model_lines = str(model).splitlines()
            for line_no in range(len(model_lines)):
                model_lines[line_no] = (
                    (branch if line_no == 0 else ("│   " if branch == "├── " else "    "))
                    + model_lines[line_no]
                    + (
                        f" | prob={self.selection_probabilities[i]:.1%}"
                        if line_no == 0
                        else ""
                    )
                )
            model_strs.append("\n".join(model_lines))

        mixer_str = (
            f"Mixer LM: {super().__str__()}[{len(self.models)}]\n"
            + "\n".join(model_strs)
        )
        return mixer_str



import multiprocessing
import os
import threading
import time
from typing import List, Tuple

from deep_translator import GoogleTranslator
from deep_translator.exceptions import NotValidPayload
from tqdm.auto import tqdm
from urllib3.exceptions import MaxRetryError, SSLError


def translate_text(text: str, gt: GoogleTranslator, results):

    try:
        translation = gt.translate(text)
    except NotValidPayload:
        translation = ""
    except (MaxRetryError, SSLError) as e:
        print(f"Error - ''{text}''")
        print(e)
        exit()

    results[text] = translation


def translate_file(data: Tuple[str, str, float]):
    """Translate a newline seperated list of words in a file into a target language using Google Translator. The source language is automatically detected.
    The output will be in a new file named "{target_language}_from_{filename}.txt"

    Args:
        data (Tuple[str, str, float]): input_file_path, target_language_code, time_delay: The amount of time in seconds to wait before sending next request in a new thread.
    """
    texts_dir, lang, time_delay = data

    with open(texts_dir, "r") as f:
        texts = f.read().splitlines()

    gt = GoogleTranslator(source="auto", target=lang)
    results = {}

    threads = []
    for txt in texts:
        thread = threading.Thread(target=translate_text, args=(txt, gt, results))
        thread.setName(f"{lang}_{txt}")
        thread.start()
        threads.append(thread)
        time.sleep(time_delay)

    for thread in threads:
        thread.join()

    translations = [results.get(txt, "") for txt in texts]

    parent, fname = os.path.split(texts_dir)
    target_dir = os.path.join(parent, f"{lang}_from_{fname}")
    with open(target_dir, "w") as f:
        f.write("\n".join(map(str, translations)))


def make_translations(args_list: List[Tuple[str, str, float]], n_processes: int = 10):
    with multiprocessing.Pool(processes=n_processes) as pool:
        for _ in tqdm(
            pool.imap_unordered(translate_file, args_list),
            total=len(args_list),
        ):
            pass



"""download from S3 or read from disk: video, landmarks [and text] data
"""

import json

import moviepy.editor as mpy
import numpy as np
import pandas as pd

from .sign_data_attributes import SignDataAttributes


class DataLoader:
    # re init
    def __init__(
        self,
        words,
        person_nos=["*"],
        camera_angles=["*"],
        signs_collection="*",
        landmarks_type="all_world_landmarks",
        dataset_directory=None,
    ) -> None:

        self.attributes = SignDataAttributes(
            words,
            person_numbers=person_nos,
            camera_angles=camera_angles,
            signs_collection=signs_collection,
            landmarks_type=landmarks_type,
            dataset_directory=dataset_directory,
        )

    def load_videos(self):
        return [mpy.VideoFileClip(pth) for pth in self.attributes.video_file_paths]

    def load_landmarks(self, reshape=False, drop_visibility=False, concat_cols=None):

        # load data by chuncks
        chunksize = 300

        chunks = pd.read_csv(
            self.attributes.landmarks_file_path,
            usecols=["Word", "Person_Number", "Camera_Angle"]
            + self.attributes.landmarks_type,
            iterator=True,
            chunksize=chunksize,
        )

        df = pd.concat(
            [
                chunk[
                    (
                        chunk["Word"].isin(self.attributes.words)
                        if "*" not in self.attributes.words
                        else np.repeat(True, len(chunk))
                    )
                    & (
                        chunk["Person_Number"].isin(self.attributes.person_nos)
                        if "*" not in self.attributes.person_nos
                        else np.repeat(True, len(chunk))
                    )
                    & (
                        chunk["Camera_Angle"].isin(self.attributes.camera_angles)
                        if "*" not in self.attributes.camera_angles
                        else np.repeat(True, len(chunk))
                    )
                ]
                for chunk in chunks
            ]
        )

        if drop_visibility and not reshape:
            reshape = True
            undo_reshape = True
        else:
            undo_reshape = False

        for land_type in self.attributes.landmarks_type:
            landmark_converter_pipeline = [lambda x: np.array(json.loads(x))]

            if reshape:
                shape = (-1, 33, 4) if land_type.startswith("Pose_") else (-1, 21, 3)
                landmark_converter_pipeline += [lambda x: x.reshape(shape)]

            if drop_visibility:
                landmark_converter_pipeline += [lambda x: x[:, :, :3]]

            if undo_reshape:
                shape = (-1, 33 * 3) if land_type.startswith("Pose_") else (-1, 21 * 3)
                landmark_converter_pipeline += [lambda x: x.reshape(shape)]

            def landmark_converter(x):
                for lc in landmark_converter_pipeline:
                    x = lc(x)
                return x

            df[land_type] = df[land_type].apply(landmark_converter)

        if concat_cols:
            new_cols = []
            removed_cols = []

            for dst_col, src_cols in concat_cols.items():
                df[dst_col] = df.apply(
                    lambda row: np.concatenate([row[col] for col in src_cols], axis=1),
                    axis=1,
                )

                new_cols += [dst_col]
                removed_cols += src_cols

            df = df[
                [col for col in list(df.columns) + new_cols if col not in removed_cols]
            ]

        return df



import re
from typing import Dict, Iterable, List

from sign_language_translator.text.utils import ListRegex


class SignTokenizer:
    def __init__(
        self,
        word_regex: str = r"\w+",
        compound_words: Iterable[str] = (),
        end_of_sentence_tokens: Iterable[str] = (".", "?", "!"),
        full_stops=(".",),
        non_sentence_end_words: Iterable[str] = ("A", "B", "C"),
        tokenized_word_sense_pattern: List|None=None,
    ):
        self.word_regex = word_regex

        self._first_subword_to_full = self._make_compound_word_map(compound_words)

        self.end_of_sentence_tokens = list(
            set(end_of_sentence_tokens) | set(full_stops)
        )
        self.non_sentence_end_words = non_sentence_end_words
        self.full_stops = full_stops
        self.tokenized_word_sense_pattern = tokenized_word_sense_pattern or [
            r"\w+",
            r"\(",
            r"\w+",
            ([r"-", r"\w+"], (0, None)),  # interval quantifier
            r"\)",
        ]

    def tokenize(
        self, text: str, join_compound_words: bool = True, join_word_sense: bool = False
    ) -> List[str]:
        matches = re.finditer(self.word_regex, text)
        word_spans = [m.span() for m in matches]

        split_indexes = []

        prev_end = -1
        for span in word_spans:
            # break up non-words into single characters
            split_indexes.extend(range(prev_end + 1, span[0]))

            split_indexes.extend(span)
            prev_end = split_indexes[-1]

        split_indexes.extend(range(prev_end + 1, len(text) + 1))

        broken = [
            text[split_indexes[i] : split_indexes[i + 1]]
            for i in range(len(split_indexes) - 1)
            if split_indexes[i] != split_indexes[i + 1]
        ]

        if join_compound_words:
            broken = self._join_subwords(broken)

        if join_word_sense:
            broken = self._join_word_sense(broken)

        return broken

    def sentence_tokenize(self, text: str) -> List[str]:
        tokens = self.tokenize(text)
        sentences = []
        sentence = []
        previous_token = None
        ended = False
        for token in tokens:
            if token in self.end_of_sentence_tokens:
                ended = True
                if (
                    token in self.full_stops
                    and previous_token in self.non_sentence_end_words  # type: ignore
                ):
                    ended = False
            else:
                if ended:
                    sentences.append(self.detokenize(sentence))
                    sentence = []
                    ended = False
            sentence.append(token)
            previous_token = token

        if sentence:
            sentences.append(self.detokenize(sentence))

        return sentences

    def detokenize(self, tokens: Iterable[str]) -> str:
        return "".join(tokens)

    def _make_compound_word_map(
        self, word_list: Iterable[str]
    ) -> Dict[str, List[List[str]]]:
        mapper: Dict[str, List] = {}
        for word in set(word_list):
            subwords = self.tokenize(word, join_compound_words=False)
            if len(subwords) > 1:
                mapper.setdefault(subwords[0], [])
                mapper[subwords[0]].append(subwords)

        # sort descending by length so that longest sequence gets joint
        for first_subword in list(mapper.keys()):
            mapper[first_subword] = sorted(mapper[first_subword], key=len, reverse=True)

        return mapper

    def _join_subwords(self, tokens: List[str]) -> List[str]:
        new_tokens = []
        i = 0
        while i < len(tokens):
            if tokens[i] in self._first_subword_to_full:
                compound_found = False
                for compound in self._first_subword_to_full[tokens[i]]:
                    if tokens[i : i + len(compound)] == compound:
                        new_tokens.append(self.detokenize(compound))
                        i += len(compound)
                        compound_found = True
                        break
                if compound_found:
                    continue

            new_tokens.append(tokens[i])
            i += 1

        return new_tokens

    def _join_word_sense(self, tokens: List[str]):
        spans = ListRegex.find_all_spans(tokens, self.tokenized_word_sense_pattern)
        spans = [(None, 0)] + spans + [(len(tokens), None)]
        new_tokens = []
        for i in range(len(spans) - 1):
            prev_end = spans[i][1]
            start = spans[i + 1][0]
            end = spans[i + 1][1]

            new_tokens.extend(tokens[prev_end:start])
            if end is not None:
                new_tokens.append("".join(tokens[start:end]))

        return new_tokens



"""utility functions for language objects"""

from __future__ import annotations

from typing import TYPE_CHECKING, Dict, Set, Type

from sign_language_translator.languages.sign import PakistanSignLanguage
from sign_language_translator.languages.text import Urdu

if TYPE_CHECKING:
    from sign_language_translator.languages.sign import SignLanguage
    from sign_language_translator.languages.text import TextLanguage


def __search_in_values_to_retrieve_key(
    code_name: str, class_to_codes: Dict[Type, Set[str]]
):
    # verify there is no repetition/reuse in language codes
    all_codes = [code for codes in class_to_codes.values() for code in codes]
    assert len(all_codes) == len(set(all_codes)), "code reused for multiple keys"

    for class_, codes in class_to_codes.items():
        if code_name.lower() in codes:
            return class_

    return None


def get_text_language(language_name: str) -> TextLanguage:
    """
    Retrieves a TextLanguage object based on the provided language name.

    Args:
        language_name (str): The name of the language.

    Returns:
        TextLanguage: An instance of the TextLanguage class corresponding to the provided language name.

    Raises:
        ValueError: If no TextLanguage class is known for the provided language name.
    """

    language_codes = {
        Urdu: {"urdu", "ur"},
    }

    class_ = __search_in_values_to_retrieve_key(language_name, language_codes)
    if class_:
        return class_()  # constructor called

    # Unknown
    raise ValueError(f"no text language class known for '{language_name = }'")


def get_sign_language(language_name: str) -> SignLanguage:
    """
    Retrieves a SignLanguage object based on the provided language name.

    Args:
        language_name (str): The name of the language.

    Returns:
        SignLanguage: An instance of SignLanguage class corresponding to the provided language name.

    Raises:
        ValueError: If no SignLanguage class is known for the provided language name.
    """

    language_codes = {
        PakistanSignLanguage: {"pakistansignlanguage", "pakistan_sign_language", "psl"},
    }

    class_ = __search_in_values_to_retrieve_key(language_name, language_codes)
    if class_:
        return class_()  # constructor called

    # Unknown
    raise ValueError(f"no sign language class known for '{language_name = }'")


__all__ = [
    "get_text_language",
    "get_sign_language",
]



from sign_language_translator.text import metrics
from sign_language_translator.text.preprocess import (
    remove_space_before_punctuation,
    replace_words,
)
from sign_language_translator.text.tagger import Rule, Tagger, Tags
from sign_language_translator.text.tokenizer import SignTokenizer

__all__ = [
    "Rule",
    "Tagger",
    "Tags",
    "SignTokenizer",
    "replace_words",
    "remove_space_before_punctuation",
    "metrics",
]



"""contains abstract class for sign languages which map spoken language text to signs using rules"""

import enum
from abc import ABC, abstractmethod
from typing import Any, Dict, Iterable, List, Tuple


class SignLanguage(ABC):
    """This abstract class defines the structure and methods required for mapping
    spoken language text to signs in sign languages using rule-based approaches.

    Attributes:
        Keys (enum.Enum): Enumerates all keys that are used in a sign dict.

    Methods:
        tokens_to_sign_dicts: Converts tokens to signs based on rules and returns a list of sign dictionaries.
        restructure_sentence: Restructures a sentence by adjusting grammar, dropping meaningless words, and normalizing synonyms.
        _make_equal_weight_sign_dict: Creates a sign dictionary with equal weights for the provided signs.
    """

    class SignDictKeys(enum.Enum):
        """Enumerates all keys that are used in a sign dict.

        Attributes:
            SIGNS (str): key for the 'signs' field in the sign dict mapping to list of sequence of video names.
            WEIGHTS (str): key for the 'weights' field in the sign dict mapping to the usage frequency of a video sequence.
        """

        SIGNS = "signs"
        WEIGHTS = "weights"

    @abstractmethod
    def tokens_to_sign_dicts(
        self,
        tokens: Iterable[str],
        tags: Iterable[Any] | None = None,
        contexts: Iterable[Any] | None = None,
    ) -> List[Dict[str, List[List[str]] | List[float]]]:
        """Converts tokens to signs based on rules and returns a list of sign dictionaries.

        Args:
            tokens (Iterable[str]): Input tokens to be converted to signs.
            tags (Iterable[Any], optional): Additional tags associated with the tokens. Defaults to None.
            contexts (Iterable[Any], optional): Additional contexts associated with the tokens. Defaults to None.

        Returns:
            List[Dict[str, List[List[str]] | List[float]]]: A list of sign dictionaries, where each dictionary contains
                the 'signs' field mapping to a list of sign sequences and the 'weights' field mapping to the usage
                frequency of each sign sequence.
                e.g. "word" -> [{"signs": [[sign_1, sign_2], [alternate_1]], "weights": [10, 5]}, ...]
        """

    @abstractmethod
    def restructure_sentence(
        self,
        sentence: Iterable[str],
        tags: Iterable[Any] | None = None,
        contexts: Iterable[Any] | None = None,
    ) -> Tuple[Iterable[str], Iterable[Any], Iterable[Any]]:
        """Restructures a sentence by changing the grammar,
        removing stopwords, spaces & punctuation, and modifying token contents.

        Args:
            sentence (Iterable[str]): Input sentence to be restructured.
            tags (Iterable[Any], optional): Additional tags associated with the sentence. Defaults to None.
            contexts (Iterable[Any], optional): Additional contexts associated with the sentence. Defaults to None.

        Returns:
            Tuple[Iterable[str], Iterable[Any], Iterable[Any]]: The restructured sentence, associated tags, and contexts.
        """

    def _make_equal_weight_sign_dict(
        self, signs: List[List[str]]
    ) -> Dict[str, List[List[str]] | List[float]]:
        """Creates a sign dictionary with equal weights for the provided signs.

        Args:
            signs (List[List[str]]): List of sign sequences.

        Returns:
            Dict[str, List[List[str]] | List[float]]: A sign dictionary with equal weights for each sign sequence.
        """

        return {
            self.SignDictKeys.SIGNS.value: signs,
            self.SignDictKeys.WEIGHTS.value: [1 / len(signs) for _ in signs],
        }



"""Extract landmarks from video
"""

import os
from typing import Any

import cv2
import mediapipe as mp
import moviepy.editor as mpy
import numpy as np


class MediapipeEmbeddingModel:
    result_keys = [
        "pose_image_landmarks",
        "left_hand_image_landmarks",
        "right_hand_image_landmarks",
        "pose_world_landmarks",
        "left_hand_world_landmarks",
        "right_hand_world_landmarks",
    ]

    def __init__(
        self,
        hands_model_complexity: int = 1,
        pose_model_complexity: int = 2,
        pose_min_detection_confidence: float = 0.5,
        pose_min_tracking_confidence: float = 0.5,
        hands_min_detection_confidence: float = 0.5,
        hands_min_tracking_confidence: float = 0.5,
        max_num_hands: int = 2,
    ) -> None:
        self.pose_model_complexity = pose_model_complexity
        self.pose_min_detection_confidence = pose_min_detection_confidence
        self.pose_min_tracking_confidence = pose_min_tracking_confidence
        self.hands_model_complexity = hands_model_complexity
        self.hands_min_detection_confidence = hands_min_detection_confidence
        self.hands_min_tracking_confidence = hands_min_tracking_confidence
        self.max_num_hands = max_num_hands

        # Mediapipe Solutions
        self.pose = mp.solutions.pose.Pose(
            model_complexity=self.pose_model_complexity,
            min_detection_confidence=self.pose_min_detection_confidence,
            min_tracking_confidence=self.pose_min_tracking_confidence,
        )
        self.hands = mp.solutions.hands.Hands(
            model_complexity=self.hands_model_complexity,
            min_detection_confidence=self.hands_min_detection_confidence,
            min_tracking_confidence=self.hands_min_tracking_confidence,
            max_num_hands=self.max_num_hands,
        )

    def mplandmark_to_nparray(landmarks, has_visibility):
        return np.array(
            [
                ([l.x, l.y, l.z, l.visibility] if has_visibility else [l.x, l.y, l.z])
                for l in landmarks.landmark
            ]
        )

    def seperate_mp_hand_landmarks(multi_hand_landmarks, multi_handedness, is_mirrored):
        landmarks = {"left": None, "right": None}

        if multi_hand_landmarks:
            for hand_landmarks, handedness in zip(
                multi_hand_landmarks, multi_handedness
            ):
                # confidence value????
                # selfie mirror flip????
                if handedness.classification[0].label == "Left":
                    landmarks["left"] = hand_landmarks
                else:
                    landmarks["right"] = hand_landmarks

        return landmarks

    # process a video frame with mediapipe
    def extract_features(self, frame, is_mirrored=False):
        """
        Extracts the pose & hands landmarks from a frame of a video clip.
        Arguments:
            frame - A Numpy array of shape (None, None, 3) representing an image from a video clip containing a person.
        Returns:
            A Tuple of Numpy arrays of shapes (33, 4), (33, 4), (42, 3), (42, 3)
            where each row is a landmark (x, y, z, [visibility]).
            First 2 items in the tuple are pose vectors (image & world)
            followed by 2 hand vectors (image & world) each with first 21 landmarks for left hand and next 21 for right hand.
        """

        landmarks = {key: None for key in MediapipeEmbeddingModel.result_keys}

        # Make Detections
        pose_results = self.pose.process(frame)
        # if is_mirrored:
        #     frame = cv2.flip(frame, 1)
        hand_results = self.hands.process(frame)

        # Extract body Landmarks
        for pose_landmarks, label in [
            (pose_results.pose_landmarks, "image"),
            (pose_results.pose_world_landmarks, "world"),
        ]:
            landmarks[f"pose_{label}_landmarks"] = (
                self.mplandmark_to_nparray(pose_landmarks, has_visibility=True)
                if pose_landmarks
                else np.zeros((33, 4))
            )

        # Extract Hand Landmarks
        for hand_landmarks, label in [
            (hand_results.multi_hand_landmarks, "image"),
            (hand_results.multi_hand_world_landmarks, "world"),
        ]:
            hand_landmarks = self.seperate_mp_hand_landmarks(
                hand_landmarks, hand_results.multi_handedness, is_mirrored
            )
            for handedness in ["left", "right"]:
                landmarks[f"{handedness}_hand_{label}_landmarks"] = (
                    self.mplandmark_to_nparray(
                        hand_landmarks[handedness], has_visibility=False
                    )
                    if hand_landmarks[handedness]
                    else np.zeros((21, 3))
                )

        return landmarks


class SignEmbedding:
    supported_embedding_models = ["mediapipe"]

    def __init__(self, embedding_model="mediapipe", *args: Any, **kwds: Any) -> None:
        if embedding_model in SignEmbedding.supported_embedding_models:
            self.EMBEDDING_MODEL_NAME = embedding_model
            self.load_embedding_model(*args, **kwds)
        else:
            raise NotImplementedError(
                f"the provided embedding model is not supported yet. use from {SignEmbedding.supported_embedding_models}"
            )

    def load_embedding_model(self, *args: Any, **kwargs: Any) -> None:
        if self.EMBEDDING_MODEL_NAME == "mediapipe":
            self.model = MediapipeEmbeddingModel(*args, **kwargs)

    def __call__(self, video, is_mirrored=False):
        return self.extract_embedding(video, is_mirrored=is_mirrored)

    def extract_embedding(self, video, is_mirrored=False):
        if isinstance(video, str):
            video_iterator = self.cv2_videofile_iterator(video)

        elif isinstance(video, mpy.VideoFileClip):
            video_iterator = video.iter_frames()

        # detect np image(s) better
        elif isinstance(video, np.ndarray):
            if (video.ndim == 3) and (3 in video.shape):
                video_iterator = video[np.newaxis, ...]
            elif (video.ndim == 4) and (3 in video.shape):
                video_iterator = video
            else:
                raise ValueError("provide 3 channel numpy image(s)")
        else:
            raise NotImplementedError(
                "provide a str filepath or a moviepy.VideoFileClip object or numpy image(s) array"
            )

        return self.extract_embedding_from_video(
            video_iterator, is_mirrored=is_mirrored
        )

    def cv2_videofile_iterator(self, video_path):
        if not os.path.exists(video_path):
            raise FileNotFoundError(f"no file exists at {video_path}")

        cap = cv2.VideoCapture(video_path)
        for frame_number in range(int(cap.get(cv2.CAP_PROP_FRAME_COUNT))):
            yield cv2.cvtColor(cap.read()[1], cv2.COLOR_BGR2RGB)
        cap.release()

    def extract_embedding_from_video(self, video, is_mirrored=False):
        """
        Extracts the pose landmarks from all frames of a video clip.
        Arguments:
            videoPath - Path of video file to be processed.
        Returns:
            Tuple of Numpy arrays of shapes (None, 33, 4), (None, 33, 4), (None, 42, 4), (None, 42, 4)
            where each row is a landmark (x, y, z, visibility).
            First 2 items in the tuple are pose vectors (image & world)
            followed by 2 hand vectors (image & world) each with 21 landmarks for each hand.
        """

        landmarks = {key: [] for key in self.model.result_keys}

        for frame in video:
            frame_landmarks = self.model.extract_features(
                frame,
                is_mirrored=is_mirrored,
            )
            for key in self.model.result_keys:
                landmarks[key].append(frame_landmarks[key])

        return {key: np.stack(_landmarks) for key, _landmarks in landmarks}



"""Defines a class for constructing Pakistan Sign Language from text using rules."""

import random
import re
from typing import Any, Dict, Iterable, List, Tuple

from sign_language_translator.config.settings import Settings
from sign_language_translator.languages.sign.mapping_rules import (
    CharacterByCharacterMappingRule,
    DirectMappingRule,
    LambdaMappingRule,
    MappingRule,
)
from sign_language_translator.languages.sign.sign_language import SignLanguage
from sign_language_translator.languages.vocab import Vocab
from sign_language_translator.text import Tags


class PakistanSignLanguage(SignLanguage):
    """A class representing the Pakistan Sign Language.

    It provides methods for converting tokens to sign dictionaries and restructuring sentences.

    Attributes:
        STOPWORDS (set): A set of stopwords in Pakistan Sign Language.
    """

    STOPWORDS = {"is", "am", "are"}

    def __init__(self) -> None:
        # load word maps and info from dataset files
        self.vocab = Vocab(
            language=r"^(?!.*\bcomponents\b).*$",  # r".*",
            sign_collections=[r"pk-hfad(-\d+)?"],  # [r"pk-.+"]
            data_root_dir=Settings.DATASET_ROOT_DIRECTORY,
            arg_is_regex=True,
        )
        # restructure dict values
        self.word_to_sign_dict = {
            word: self._make_equal_weight_sign_dict(labels)
            for word, labels in self.vocab.word_to_labels.items()
        }

        # define mapping rules
        self._direct_rule = self.__get_direct_mapping_rule(4)
        self._double_handed_spelling_rule = self.__get_spelling_rule(5, "double-handed")
        self._single_handed_spelling_rule = self.__get_spelling_rule(5, "single-handed")
        self._urdu_character_rule = self.__get_urdu_spelling_rule(5)
        self._number_rule = self.__get_number_rule(5)

        self.mapping_rules: List[MappingRule] = [
            self._direct_rule,
            self._double_handed_spelling_rule,
            self._single_handed_spelling_rule,
            self._urdu_character_rule,
            self._number_rule,
            # self._time_rule, # x:y:z -> x hours y minutes z seconds
            # self._date_rule,
        ]

    def tokens_to_sign_dicts(
        self,
        tokens: Iterable[str],
        tags: Iterable[Any] | None = None,
        contexts: Iterable[Any] | None = None,
    ) -> List[Dict[str, List[List[str]] | List[float]]]:
        # fix args
        if isinstance(tokens, str):
            tokens = [tokens]
        if not tags:
            tags = [None for _ in tokens]
        if not contexts:
            contexts = [None for _ in tokens]

        # map tokens to signs
        signs = [
            sign_dict
            for token, tag, context in zip(tokens, tags, contexts)
            for sign_dict in self._apply_rules(token, tag, context)
        ]

        return signs

    def _apply_rules(
        self, token: str, tag=None, context=None
    ) -> List[Dict[str, List[List[str]] | List[float]]]:
        """Applies all the mapping rules to a token.
        Rules with lower value of priority overwrite the result.
        If multiple rules of same priority are applicable, one is selected at random.

        Args:
            token (str): The token to apply the rules to.
            tag (Any, optional): The tag associated with the token. Defaults to None.
            context (Any, optional): The context associated with the token. Defaults to None.

        Returns:
            List[Dict[str, List[List[str]] | List[float]]]:
                A list of sign dictionaries for the token.
        """

        sign = None
        priority = float("inf")

        for rule in self.mapping_rules:
            if rule.is_applicable(token, tag, context):
                if rule.priority < priority or (
                    rule.priority == priority and random.random() < 0.5
                ):
                    sign = rule.apply(token)
                    priority = rule.priority

        if sign is None:
            raise ValueError(f"No PakistanSL sign could be inferred for {token = }")

        return sign

    def restructure_sentence(
        self,
        sentence: Iterable[str],
        tags: Iterable[Any] | None = None,
        contexts: Iterable[Any] | None = None,
    ) -> Tuple[Iterable[str], Iterable[Any], Iterable[Any]]:
        # Fix the args
        tags = [Tags.DEFAULT for _ in sentence] if tags is None else tags
        contexts = [None for _ in sentence] if contexts is None else contexts

        # map to urdu grammar "he goes to school" -> ["he", "school", "go"]
        restructured_sentence = []
        restructured_tags = []
        restructured_contexts = []

        # drop stuff
        for token, tag, context in zip(sentence, tags, contexts):
            # drop stop-words
            if token in self.STOPWORDS:
                continue

            # drop space and punctuation
            if tag in {Tags.SPACE, Tags.PUNCTUATION}:
                continue

            # make numbers "numeric"
            if tag == Tags.NUMBER and "," in token:
                token = token.replace(",", "")

            # drop word-sense: "(name)" from the NAME token
            if tag == Tags.NAME:
                token = re.sub(self.vocab.word_sense_regex, "", token)

            restructured_sentence.append(token)
            restructured_tags.append(tag)
            restructured_contexts.append(context)

        return restructured_sentence, restructured_tags, restructured_contexts

    def __call__(
        self,
        tokens: Iterable[str],
        tags: Iterable[Any] | None = None,
        contexts: Iterable[Any] | None = None,
    ) -> List[Dict[str, List[List[str]] | List[float]]]:
        tokens, tags, contexts = self.restructure_sentence(
            tokens, tags=tags, contexts=contexts
        )
        signs = self.tokens_to_sign_dicts(tokens, tags=tags, contexts=contexts)

        return signs

    def __get_direct_mapping_rule(self, priority=5):
        return DirectMappingRule(
            {w: [sd] for w, sd in self.word_to_sign_dict.items()}, priority
        )

    def __get_spelling_rule(
        self,
        priority: int,
        word_sense_filter="handed-letter",
    ):
        return CharacterByCharacterMappingRule(
            {
                k__: v
                for k, v in self.word_to_sign_dict.items()
                for k_ in [self.vocab.remove_word_sense(k)]
                for k__ in [k_.lower(), k_.upper()]
                if word_sense_filter in k
            },
            {Tags.NAME, Tags.ACRONYM},
            priority,
        )

    def __get_urdu_spelling_rule(self, priority=5):
        return CharacterByCharacterMappingRule(
            {
                k: v
                for k, v in self.word_to_sign_dict.items()
                if len(k) == 1 and not k.isnumeric()
            },
            {Tags.NAME},
            priority,
        )

    def __get_number_rule(self, priority=5):
        return LambdaMappingRule(
            is_applicable_function=lambda token, tag, context: (
                tag == Tags.NUMBER
                and all(
                    digit in self.word_to_sign_dict
                    for digit in self.__chunk_a_number(token)
                )
            ),
            apply_function=lambda x: [
                self.word_to_sign_dict[digit] for digit in self.__chunk_a_number(x)
            ],
            priority=priority,
        )

    def __chunk_a_number(self, num):
        return re.findall(
            r"("
            + r"|".join(sorted(self.vocab.numeric_keys, key=len, reverse=True))
            + r"|\d|\.|.)",
            str(self.vocab.words_to_numbers.get(num, num)),
        )



"""Toolkit for ploting pose vectors

Bugs:
    - too much white space around the figures
"""

from typing import Dict, List, Tuple, Union

import matplotlib
import matplotlib.pyplot as plt
import numpy as np
import torch
from moviepy.video.io.bindings import mplfig_to_npimage
from moviepy.video.VideoClip import DataVideoClip
from PIL.Image import fromarray as np2pil
import PIL.Image

from ..utils.landmarks_info import LandmarksInfo


def _get_new_plt_fig(
    axis_lims: List[float] = None,
    elev=15,
    azim=25,
    fig_width=10,
    fig_height=10,
    box_aspect=[1.0, 1.0, 1.0],
) -> Tuple[matplotlib.figure.Figure, matplotlib.axes.Axes]:
    """Initialize an empty 3D graph

    Args:
        axis_lims (List[float], optional): the upper and lower limits of the 3D plot axis [x_lower, x_upper, y_lower, y_upper, z_lower, z_upper]. Defaults to None.
        elev (int, optional): the elevation angle of view point in degrees. Defaults to 15.
        azim (int, optional): the azimuth angle of view point in degrees. Defaults to 25.
        fig_width (int, optional): the width of the graph in inches. Defaults to 10.
        fig_height (int, optional): the height of the graph in inches. Defaults to 10.
        box_aspect (List[float], optional): the aspect ratio of x, y, z axes

    Returns:
        Tuple[matplotlib.figure.Figure, matplotlib.axes.Axes]: an empty matplotlib (fig, ax) pair
    """

    fig = plt.figure(figsize=(fig_width, fig_height))
    ax = fig.add_subplot(projection="3d")

    ax.set_box_aspect(box_aspect)
    ax.view_init(elev=elev, azim=azim, vertical_axis="y")

    if axis_lims:
        ax.set_xlim(axis_lims[0], axis_lims[1])
        ax.set_ylim(axis_lims[2], axis_lims[3])
        ax.set_zlim(axis_lims[4], axis_lims[5])

    ax.set_xlabel("X")
    ax.set_ylabel("Y")
    ax.set_zlabel("Z")

    fig.tight_layout()

    ax.invert_zaxis()
    ax.invert_yaxis()

    return fig, ax


# put landmarks on  3D graph
def plot_landmarks(
    landmarks: np.ndarray,
    connections: List[List[int]] = None,
    fig: matplotlib.figure.Figure = None,
    ax: matplotlib.axes.Axes = None,
    axis_lims: List[float] = None,
    box_aspect=[1.0, 1.0, 1.0],
    elev=15,
    azim=25,
    landmarks_color: Tuple[float, float, float] = LandmarksInfo.Colors.DEFAULT_LANDMARK,
    landmark_size=6.9,
    alpha=1.0,
    return_plot_cache=False,
    fig_width=10,
    fig_height=10,
) -> Union[
    Tuple[
        matplotlib.figure.Figure,
        matplotlib.axes.Axes,
        Dict[str, object],
    ],
    Tuple[matplotlib.figure.Figure, matplotlib.axes.Axes],
]:
    """Plots a single pose & hands vector in 3D.

    Args:
        landmarks (np.ndarray): A numpy array of dimention (n,3) containing n landmarks (x, y, z).
        connections (List[List[int]], optional): Indices of the landmarks array to be connected by a line. Defaults to None.
        fig (matplotlib.figure.Figure, optional): matplotlib figure. Defaults to None.
        ax (matplotlib.axes.Axes, optional): matplotlib axes. Defaults to None.
        axis_lims (List[float], optional): the upper and lower limits of the 3D plot axis [x_lower, x_upper, y_lower, y_upper, z_lower, z_upper]. Defaults to None.
        box_aspect (List[float], optional): the aspect ratio of x, y, z axes
        elev (int, optional): Elevation angle of view point in degrees. Defaults to 15.
        azim (int, optional): Azimuth angle of view point in degrees. Defaults to 25.
        landmarks_color (Tuple[float,float,float], optional): the color as RGB values [0 to 1] of the dots representing landmarks. Defaults to LandmarksInfo.Colors.DEFAULT_LANDMARK.
        landmark_size (float, optional): Size of the dots drawn. Defaults to 6.9.
        alpha (float, optional): Opacity of the landmark. Defaults to 1.0.
        return_plot_cache (bool, optional): Wether to return the a dict of objects corresponding to lines and dots drawn. Allows the data to be updated. Defaults to False.
        fig_width (int, optional): the width of the graph in inches. Defaults to 10.
        fig_height (int, optional): the height of the graph in inches. Defaults to 10.

    Returns:
        Tuple[matplotlib.figure.Figure, matplotlib.axes.Axes [, Dict[str, object]]] : matplotlib fig, ax [, plots_cache]
    """

    landmarks = LandmarksInfo.Reshaper().fold(landmarks)

    if isinstance(landmarks, torch.Tensor):
        landmarks = landmarks.numpy()

    # set up a matplotlib plot
    if fig is None or ax is None:
        fig, ax = _get_new_plt_fig(
            axis_lims=axis_lims,
            elev=elev,
            azim=azim,
            box_aspect=box_aspect,
            fig_height=fig_height,
            fig_width=fig_width,
        )

    if connections is None:
        connections = LandmarksInfo.infer_connections(len(landmarks))

    get_connection_color = (
        LandmarksInfo.Colors._get_hand_connection_color
        if connections == LandmarksInfo.HAND_CONNECTIONS
        else LandmarksInfo.Colors._get_pose_connection_color
        if connections == LandmarksInfo.POSE_CONNECTIONS
        else LandmarksInfo.Colors._get_both_hands_connection_color
        if connections == LandmarksInfo.BOTH_HANDS_CONNECTIONS
        else LandmarksInfo.Colors._get_all_connection_color
        if connections == LandmarksInfo.ALL_CONNECTIONS
        else LandmarksInfo.Colors._get_default_connection_color
    )

    def infer_label(color=(0.0, 0.0, 0.0)) -> str:
        return (
            "Left"
            if color == LandmarksInfo.Colors.LEFT
            else "Right"
            if color == LandmarksInfo.Colors.RIGHT
            else "Thumb"
            if color == LandmarksInfo.Colors.THUMB
            else "Pinky"
            if color == LandmarksInfo.Colors.PINKY
            else None
        )

    plot_cache = {"landmarks_shape": landmarks.shape, "connection_lines": dict()}

    # plot landmarks as dots
    plot_cache["scatter3D"] = ax.scatter3D(
        xs=landmarks[:, 0],
        ys=landmarks[:, 1],
        zs=landmarks[:, 2],
        color=landmarks_color,
        s=landmark_size,
        alpha=alpha,
    )

    # draw lines to connect landmarks
    for connection in connections:
        start, end = connection

        color = get_connection_color(start, end)
        label = infer_label(color)

        plot_cache["connection_lines"][(start, end)] = ax.plot3D(
            xs=[landmarks[start, 0], landmarks[end, 0]],
            ys=[landmarks[start, 1], landmarks[end, 1]],
            zs=[landmarks[start, 2], landmarks[end, 2]],
            color=color,
            linewidth=2,
            label=label,
            alpha=alpha,
        )

    # legend redundancy removal
    handles, labels = plt.gca().get_legend_handles_labels()
    by_label = dict(zip(labels, handles))
    plt.legend(by_label.values(), by_label.keys())

    if return_plot_cache:
        return fig, ax, plot_cache
    else:
        return fig, ax


def update_plotted_landmarks(
    landmarks: np.ndarray, plot_cache: Dict[str, object]
) -> None:
    """change the data in an existing plot so nothing else has to be drawn again.

    Args:
        landmarks (np.ndarray): A numpy array of dimention (n,3) containing n landmarks (x, y, z).
        plot_cache (Dict[str, object]): Maps strings to lines and dots drawn so they can be moved.
    """

    if isinstance(landmarks, torch.Tensor):
        landmarks = landmarks.numpy()

    assert (
        landmarks.shape == plot_cache["landmarks_shape"]
    ), "the provided landmarks do not match the shape of the landmarks in the cache"

    plot_cache["scatter3D"]._offsets3d = (
        landmarks[:, 0],
        landmarks[:, 1],
        landmarks[:, 2],
    )

    for (start, end), line in plot_cache["connection_lines"].items():
        line[0].set_data(
            [
                [landmarks[start, 0], landmarks[end, 0]],
                [landmarks[start, 1], landmarks[end, 1]],
            ]
        )
        line[0].set_3d_properties([landmarks[start, 2], landmarks[end, 2]])


def _get_box_coord_ranges(multi_frame_landmarks: np.ndarray) -> List[float]:
    """find the range of values within which all landmarks stay in all frames

    Args:
        multi_frame_landmarks (np.ndarray): an array of dimentions ([n_frames], n_landmarks, n_axes=3)

    Returns:
        List[float]: the range of values in each dimention [x_min, x_max, y_min, y_max, z_min, z_max]
    """

    assert (
        1 < len(multi_frame_landmarks.shape) <= 3
    ), "all_landmarks should be 2 or 3 dimentional"

    return [
        np.min(multi_frame_landmarks[..., 0]),
        np.max(multi_frame_landmarks[..., 0]),
        np.min(multi_frame_landmarks[..., 1]),
        np.max(multi_frame_landmarks[..., 1]),
        np.min(multi_frame_landmarks[..., 2]),
        np.max(multi_frame_landmarks[..., 2]),
    ]


def _get_box(
    x_min: float,
    x_max: float,
    y_min: float,
    y_max: float,
    z_min: float,
    z_max: float,
    full=True,
) -> Tuple[List[List[float]], List[List[int]]]:
    """Create a square or cuboid corners coordinates and connections from range values of axes

    Args:
        x_min (float): the lowest x value
        x_max (float): the highest x value
        y_min (float): the lowest y value
        y_max (float): the highest y value
        z_min (float): the lowest z value
        z_max (float): the highest z value
        full (bool, optional): True means return connections of a cuboid and False means return only a square on y_max side. Defaults to True.

    Returns:
        Tuple[List[List[float]], List[List[int]]]: box corner coordinates, box_connections
    """

    box_coords = np.array(
        [
            [x_min, y_min, z_min],
            [x_min, y_min, z_max],
            [x_min, y_max, z_min],
            [x_min, y_max, z_max],
            [x_max, y_min, z_min],
            [x_max, y_min, z_max],
            [x_max, y_max, z_min],
            [x_max, y_max, z_max],
        ]
    )

    box_connections = (
        [
            [0, 1],
            [0, 2],
            [0, 4],
            [7, 6],
            [7, 5],
            [7, 3],
            [1, 3],
            [3, 2],
            [2, 6],
            [6, 4],
            [4, 5],
            [5, 1],
        ]
        if full
        else [
            [7, 6],
            [6, 2],
            [2, 3],
            [3, 7],
        ]
    )

    return box_coords, box_connections


def _move_hands_apart(landmarks: np.ndarray, distance=0.2, scale=1.0) -> np.ndarray:
    """updates the hand landmarks such that left hand gets a positive shift in x-dimention and right hand gets a negative shift

    Args:
        landmarks (np.ndarray): an array with 2nd last dimention containing all landmarks of 2 hands and last dimention contains [x, ...] values
        distance (float, optional): how many units to move the hands. Defaults to 0.2.
        scale (float, optional): how much to zoom the hands. Defaults to 1.0.

    Returns:
        np.ndarray: the modified version of the landmarks array
    """

    landmarks *= scale

    if landmarks.shape[-2] == LandmarksInfo.N_BOTH_HANDS_LANDMARKS:

        landmarks[..., : LandmarksInfo.N_HAND_LANDMARKS, 0] += (
            distance / 2
        )  # left  hand
        landmarks[..., LandmarksInfo.N_HAND_LANDMARKS :, 0] -= (
            distance / 2
        )  # right hand

    return landmarks


def _move_hands_to_wrists(landmarks: np.ndarray, scale=1.0) -> np.ndarray:
    """translates the hand landmarks which have origin on their geometric center to the wrists of the body landmarks

    Args:
        landmarks (np.ndarray): an array with 2nd last dimention containing all landmarks of body and 2 hands and last dimention contains [x, ...] values
        scale (float, optional): how much to zoom the hands. Defaults to 1.0.

    Returns:
        np.ndarray: the modified version of the landmarks array
    """

    if landmarks.shape[-2] == LandmarksInfo.N_ALL_LANDMARKS:

        # Left hand
        left_wrist = landmarks[
            ...,
            LandmarksInfo.POSE_LEFT_WRIST_INDEX : LandmarksInfo.POSE_LEFT_WRIST_INDEX
            + 1,
            :,
        ]
        left_hand_landmarks = landmarks[
            ...,
            LandmarksInfo.N_POSE_LANDMARKS : LandmarksInfo.N_POSE_LANDMARKS
            + LandmarksInfo.N_HAND_LANDMARKS,
            :,
        ]
        left_hand_landmarks *= scale
        left_hand_landmarks -= left_hand_landmarks[
            ..., LandmarksInfo.HAND_WRIST_INDEX : LandmarksInfo.HAND_WRIST_INDEX + 1, :
        ]
        left_hand_landmarks += left_wrist

        # Right hand
        right_wrist = landmarks[
            ...,
            LandmarksInfo.POSE_RIGHT_WRIST_INDEX : LandmarksInfo.POSE_RIGHT_WRIST_INDEX
            + 1,
            :,
        ]
        right_hand_landmarks = landmarks[
            ...,
            LandmarksInfo.N_POSE_LANDMARKS
            + LandmarksInfo.N_HAND_LANDMARKS : LandmarksInfo.N_POSE_LANDMARKS
            + 2 * LandmarksInfo.N_HAND_LANDMARKS,
            :,
        ]
        right_hand_landmarks *= scale
        right_hand_landmarks -= right_hand_landmarks[
            ..., LandmarksInfo.HAND_WRIST_INDEX : LandmarksInfo.HAND_WRIST_INDEX + 1, :
        ]
        right_hand_landmarks += right_wrist

    return landmarks


def _move_hands(
    landmarks: np.ndarray,
    connections: List[List[int]],
    hands_location: str,
    scale=1.0,
    distance=0.2,
) -> np.ndarray:
    """move origin centered hands apart or to wrists

    Args:
        landmarks (np.ndarray): A numpy array of dimention ([n_frames], n_landmarks, n_axes=3) containing (x, y, z) coordinates
        hands_location (str): target location of hands "auto", "wrists" or "spaced". None makes no change in landmarks.
        scale (float, optional): how many times should hands be zoomed. Defaults to 1.0.
        distance (float, optional): how much should they move "apart". Defaults to 0.2.

    Raises:
        ValueError: bad value of hands_location argument

    Returns:
        np.ndarray: the modified version of the landmarks array
    """

    if isinstance(landmarks, torch.Tensor):
        landmarks = landmarks.numpy()

    landmarks = landmarks.copy()

    if connections is None:
        connections = LandmarksInfo.infer_connections(landmarks.shape[-2])

    if connections == LandmarksInfo.BOTH_HANDS_CONNECTIONS:
        if hands_location is None:
            pass
        elif hands_location in ["auto", "spaced"]:
            landmarks = _move_hands_apart(landmarks, scale=scale, distance=distance)
        else:
            raise ValueError(
                'bad value of hands_location argument, should be None or "spaced" or "auto"'
            )

    elif connections == LandmarksInfo.ALL_CONNECTIONS:
        if hands_location is None:
            pass
        elif hands_location in ["auto", "wrists"]:
            landmarks = _move_hands_to_wrists(landmarks, scale=scale)
        else:
            raise ValueError(
                'bad value of hands_location argument, should be None or "wrists" or "auto"'
            )

    return landmarks


def _fig_to_npimage(fig: matplotlib.figure.Figure) -> np.ndarray:
    """convert a matplotlib figure to a numpy array

    Args:
        fig (matplotlib.figure.Figure): A matplotlib figure object

    Returns:
        np.ndarray: a 3 channel RGB image array
    """

    return mplfig_to_npimage(fig)


def landmarks_to_npimage(
    landmarks: np.ndarray,
    connections: List[List[int]] = None,
    landmarks_color: Tuple[float, float, float] = LandmarksInfo.Colors.DEFAULT_LANDMARK,
    axis_lims: List[float] = None,
    elev=15,
    azim=25,
    move_hands=True,
    hands_location="auto",
    hands_scale=1.0,
    hands_distance=0.1,
) -> np.ndarray:
    """draw single frame landmarks onto an image

    Args:
        landmarks (np.ndarray): A numpy array of dimention (n,3) containing n landmarks (x, y, z).
        connections (List[List[int]], optional): Indices of the landmarks array to be connected by a line.. Defaults to None.
        landmarks_color (Tuple[float,float,float], optional): the color as RGB values [0 to 1] of the dots representing landmarks. Defaults to LandmarksInfo.Colors.DEFAULT_LANDMARK.
        axis_lims (List[float], optional): the upper and lower limits of the 3D plot axis [x_lower, x_upper, y_lower, y_upper, z_lower, z_upper].. Defaults to None.
        elev (int, optional): the elevation angle of view point in degrees. Defaults to 15.
        azim (int, optional): Azimuth angle of view point in degrees. Defaults to 25.
        move_hands (bool, optional): wether to move origin centered hands apart or to wrists or not. Defaults to True.
        hands_location (str, optional): where to move hand to. ignored if move_hands is false. Defaults to "auto".
        hands_scale (float, optional): how much should hands be zoomed. Defaults to 1.0.
        hands_distance (float, optional): how much should hands be shifted "apart". Defaults to 0.1.

    Returns:
        np.ndarray: a 3 channel RGB image array containing drawn landmarks
    """

    landmarks = LandmarksInfo.Reshaper().fold(landmarks)

    if isinstance(landmarks, torch.Tensor):
        landmarks = landmarks.numpy()

    if connections is None:
        connections = LandmarksInfo.infer_connections(len(landmarks))

    if move_hands:
        landmarks = _move_hands(
            landmarks,
            connections,
            hands_location,
            scale=hands_scale,
            distance=hands_distance,
        )

    fig, ax = plot_landmarks(
        landmarks,
        connections,
        axis_lims=axis_lims,
        elev=elev,
        azim=azim,
        landmarks_color=landmarks_color,
    )

    body_ratios = (
        np.ptp(landmarks[:, 2]),
        np.ptp(landmarks[:, 0]),
        np.ptp(landmarks[:, 1]),
    )

    ax.set_box_aspect(body_ratios)

    image = _fig_to_npimage(fig)

    plt.clf()
    plt.close()

    return image


def plot_multi_frame_landmarks(
    multi_frame_landmarks: np.ndarray,
    connections: List[List[int]] = None,
    landmarks_color: Tuple[float, float, float] = LandmarksInfo.Colors.DEFAULT_LANDMARK,
    axis_lims: List[float] = None,
    elev=15,
    azim=25,
    move_hands=True,
    hands_location="auto",
    hands_scale=1.0,
    hands_distance=0.1,
) -> Tuple[matplotlib.figure.Figure, matplotlib.axes.Axes]:
    """Plot multiple frames on a single graph with x-shift

    Args:
        multi_frame_landmarks (np.ndarray): an array of dimentions (n_frames, n_landmarks, n_axes=3)
        connections (List[List[int]], optional): Indices of the landmarks array to be connected by a line. Defaults to None.
        landmarks_color (Tuple[float, float, float], optional): the color as RGB values [0 to 1] of the dots representing landmarks. Defaults to LandmarksInfo.Colors.DEFAULT_LANDMARK.
        axis_lims (List[float], optional): the upper and lower limits of the 3D plot axis [x_lower, x_upper, y_lower, y_upper, z_lower, z_upper].. Defaults to None.
        elev (int, optional): the elevation angle of view point in degrees. Defaults to 15.
        azim (int, optional): Azimuth angle of view point in degrees. Defaults to 25.
        move_hands (bool, optional): wether to move origin centered hands apart or to wrists or not. Defaults to True.
        hands_location (str, optional): where to move hand to. ignored if move_hands is false. Defaults to "auto".
        hands_scale (float, optional): how much should hands be zoomed. Defaults to 1.0.
        hands_distance (float, optional): how much should hands be shifted "apart". Defaults to 0.1.

    Returns:
        Tuple[matplotlib.figure.Figure, matplotlib.axes.Axes]: matplotlib fig, ax pair with larmarks plotted
    """

    multi_frame_landmarks = LandmarksInfo.Reshaper().fold(multi_frame_landmarks)

    if isinstance(multi_frame_landmarks, torch.Tensor):
        multi_frame_landmarks = multi_frame_landmarks.numpy()

    if connections is None:
        connections = LandmarksInfo.infer_connections(multi_frame_landmarks.shape[-2])

    if move_hands:
        multi_frame_landmarks = _move_hands(
            multi_frame_landmarks,
            connections,
            hands_location,
            scale=hands_scale,
            distance=hands_distance,
        )

    x0, x1, y0, y1, z0, z1 = _get_box_coord_ranges(multi_frame_landmarks)
    box_coords, box_connections = _get_box(x0, x1, y0, y1, z0, z1, full=False)

    x_shift = (x1 - x0) * 1.25

    fig, ax = _get_new_plt_fig(
        axis_lims=axis_lims,
        elev=elev,
        azim=azim,
    )  # fig_width = 10*len(multi_frame_landmarks))

    # Make graph wider
    n_frames = len(multi_frame_landmarks)
    body_ratios = (z1 - z0, x_shift * n_frames, y1 - y0)
    ax.set_box_aspect(body_ratios)

    for i, landmarks in enumerate(multi_frame_landmarks.copy()):
        box_coords[:, 0] += x_shift if i > 0 else 0
        landmarks[:, 0] += x_shift * i

        fig, ax = plot_landmarks(
            box_coords,
            box_connections,
            fig=fig,
            ax=ax,
            axis_lims=axis_lims,
            elev=elev,
            azim=azim,
            landmarks_color=LandmarksInfo.Colors._get_default_connection_color(),
            landmark_size=0,
            alpha=0.5,
        )
        fig, ax = plot_landmarks(
            landmarks,
            connections,
            fig=fig,
            ax=ax,
            axis_lims=axis_lims,
            elev=elev,
            azim=azim,
            landmarks_color=landmarks_color,
        )

    return fig, ax


def multi_frame_landmarks_to_npimage(
    multi_frame_landmarks: np.ndarray,
    connections: List[List[int]] = None,
    landmarks_color: Tuple[float, float, float] = LandmarksInfo.Colors.DEFAULT_LANDMARK,
    axis_lims: List[float] = None,
    elev=15,
    azim=25,
    move_hands=True,
    hands_location="auto",
    hands_scale=1.0,
    hands_distance=0.1,
) -> np.ndarray:
    """Draw multiple frames on a single graph image with x-shift

    Args:
        multi_frame_landmarks (np.ndarray): an array of dimentions (n_frames, n_landmarks, n_axes=3)
        connections (List[List[int]], optional): Indices of the landmarks array to be connected by a line. Defaults to None.
        landmarks_color (Tuple[float, float, float], optional): the color as RGB values [0 to 1] of the dots representing landmarks. Defaults to LandmarksInfo.Colors.DEFAULT_LANDMARK.
        axis_lims (List[float], optional): the upper and lower limits of the 3D plot axis [x_lower, x_upper, y_lower, y_upper, z_lower, z_upper].. Defaults to None.
        elev (int, optional): the elevation angle of view point in degrees. Defaults to 15.
        azim (int, optional): Azimuth angle of view point in degrees. Defaults to 25.
        move_hands (bool, optional): wether to move origin centered hands apart or to wrists or not. Defaults to True.
        hands_location (str, optional): where to move hand to. ignored if move_hands is false. Defaults to "auto".
        hands_scale (float, optional): how much should hands be zoomed. Defaults to 1.0.
        hands_distance (float, optional): how much should hands be shifted "apart". Defaults to 0.1.

    Returns:
        np.ndarray: a 3 channel RGB image array containing drawn landmarks
    """

    multi_frame_landmarks = LandmarksInfo.Reshaper().fold(multi_frame_landmarks)

    if isinstance(multi_frame_landmarks, torch.Tensor):
        multi_frame_landmarks = multi_frame_landmarks.numpy()

    fig, ax = plot_multi_frame_landmarks(
        multi_frame_landmarks,
        connections=connections,
        landmarks_color=landmarks_color,
        axis_lims=axis_lims,
        elev=elev,
        azim=azim,
        move_hands=move_hands,
        hands_location=hands_location,
        hands_scale=hands_scale,
        hands_distance=hands_distance,
    )
    image = _fig_to_npimage(fig)

    plt.clf()
    plt.close()

    return image

def multi_frame_landmarks_to_pil_image(
    multi_frame_landmarks: np.ndarray,
    connections: List[List[int]] = None,
    landmarks_color: Tuple[float, float, float] = LandmarksInfo.Colors.DEFAULT_LANDMARK,
    axis_lims: List[float] = None,
    elev=15,
    azim=25,
    move_hands=True,
    hands_location="auto",
    hands_scale=1.0,
    hands_distance=0.1,
) -> PIL.Image:
    """Draw multiple frames on a single graph image with x-shift

    Args:
        multi_frame_landmarks (np.ndarray): an array of dimentions (n_frames, n_landmarks, n_axes=3)
        connections (List[List[int]], optional): Indices of the landmarks array to be connected by a line. Defaults to None.
        landmarks_color (Tuple[float, float, float], optional): the color as RGB values [0 to 1] of the dots representing landmarks. Defaults to LandmarksInfo.Colors.DEFAULT_LANDMARK.
        axis_lims (List[float], optional): the upper and lower limits of the 3D plot axis [x_lower, x_upper, y_lower, y_upper, z_lower, z_upper].. Defaults to None.
        elev (int, optional): the elevation angle of view point in degrees. Defaults to 15.
        azim (int, optional): Azimuth angle of view point in degrees. Defaults to 25.
        move_hands (bool, optional): wether to move origin centered hands apart or to wrists or not. Defaults to True.
        hands_location (str, optional): where to move hand to. ignored if move_hands is false. Defaults to "auto".
        hands_scale (float, optional): how much should hands be zoomed. Defaults to 1.0.
        hands_distance (float, optional): how much should hands be shifted "apart". Defaults to 0.1.

    Returns:
        PIL.Image: a 3 channel RGB image array containing drawn landmarks
    """

    np_image = multi_frame_landmarks_to_npimage(
        multi_frame_landmarks = multi_frame_landmarks,
        connections=connections,
        landmarks_color=landmarks_color,
        axis_lims=axis_lims,
        elev=elev,
        azim=azim,
        move_hands=move_hands,
        hands_location=hands_location,
        hands_scale=hands_scale,
        hands_distance=hands_distance,
    )
    pil_image = np2pil(np_image)

    return pil_image


def multi_frame_landmarks_to_video(
    multi_frame_landmarks: np.ndarray,
    connections: List[List[int]] = None,
    landmarks_color: Tuple[float, float, float] = LandmarksInfo.Colors.DEFAULT_LANDMARK,
    axis_lims: List[float] = None,
    elev=15.0,
    azim=25.0,
    fps=24.0,
    rotate=True,
    move_hands=True,
    hands_location="auto",
    hands_scale=1.0,
    hands_distance=0.1,
    fig_width=10,
    fig_height=10,
) -> DataVideoClip:
    """Animate a pose vector's motion in 3D

    Args:
        multi_frame_landmarks (np.ndarray): an array of dimentions (n_frames, n_landmarks, n_axes=3)
        connections (List[List[int]], optional): Indices of the landmarks array to be connected by a line. Defaults to None.
        landmarks_color (Tuple[float, float, float], optional): the color as RGB values [0 to 1] of the dots representing landmarks. Defaults to LandmarksInfo.Colors.DEFAULT_LANDMARK.
        axis_lims (List[float], optional): the upper and lower limits of the 3D plot axis [x_lower, x_upper, y_lower, y_upper, z_lower, z_upper].. Defaults to None.
        elev (int, optional): the elevation angle of view point in degrees. Defaults to 15.
        azim (int, optional): Azimuth angle of view point in degrees. Defaults to 25.
        fps (float, optional): the video frame rate per second. Defaults to 24.0.
        rotate (bool, optional): wether to add a cinematic effect to the animation. Defaults to True.
        move_hands (bool, optional): wether to move origin centered hands apart or to wrists or not. Defaults to True.
        hands_location (str, optional): where to move hand to. ignored if move_hands is false. Defaults to "auto".
        hands_scale (float, optional): how much should hands be zoomed. Defaults to 1.0.
        hands_distance (float, optional): how much should hands be shifted "apart". Defaults to 0.1.
        fig_width (int, optional): the width of the graph in inches. Defaults to 10.
        fig_height (int, optional): the height of the graph in inches. Defaults to 10.

    Returns:
        moviepy.video.VideoClip.DataVideoClip: a video of pose vector's motion in 3D
    """

    multi_frame_landmarks = LandmarksInfo.Reshaper().fold(multi_frame_landmarks)

    if isinstance(multi_frame_landmarks, torch.Tensor):
        multi_frame_landmarks = multi_frame_landmarks.numpy()

    if connections is None:
        connections = LandmarksInfo.infer_connections(multi_frame_landmarks.shape[-2])

    if move_hands:
        multi_frame_landmarks = _move_hands(
            multi_frame_landmarks,
            connections,
            hands_location,
            scale=hands_scale,
            distance=hands_distance,
        )

    x0, x1, y0, y1, z0, z1 = _get_box_coord_ranges(multi_frame_landmarks)

    if axis_lims is None:
        scale = 0.05
        axis_lims = [
            x0 - abs(x0 * scale),
            x1 + abs(x1 * scale),
            y0 - abs(y0 * scale),
            y1 + abs(y1 * scale),
            z0 - abs(z0 * scale),
            z1 + abs(z1 * scale),
        ]

    fig, ax, plot_cache = plot_landmarks(
        multi_frame_landmarks[0],
        connections,
        axis_lims=axis_lims,
        elev=elev,
        azim=azim,
        landmarks_color=landmarks_color,
        return_plot_cache=True,
        fig_width=fig_width,
        fig_height=fig_height,
    )
    ax.set_box_aspect((z1 - z0, x1 - x0, y1 - y0))

    def data_to_frame(t):
        if rotate:
            ax.view_init(
                elev=elev - elev * t * 1 / n_frames,
                azim=azim - azim * t * 2 / n_frames,
                vertical_axis="y",
            )

        if t > 0:
            update_plotted_landmarks(multi_frame_landmarks[t], plot_cache)

        image = _fig_to_npimage(fig)

        if t == len(multi_frame_landmarks) - 1:
            plt.clf()
            plt.close()

        return image

    n_frames = len(multi_frame_landmarks)
    timesteps = list(range(n_frames))
    clip = DataVideoClip(timesteps, data_to_frame, fps=fps)

    return clip



"""This module defines the ConcatenativeSynthesis class, which represents a rule based model for translating text to sign language."""

from __future__ import annotations

import random
from typing import TYPE_CHECKING

from sign_language_translator.languages import get_sign_language, get_text_language
from sign_language_translator.models.text_to_sign.t2s_model import TextToSignModel

if TYPE_CHECKING:
    from sign_language_translator.languages.sign import SignLanguage
    from sign_language_translator.languages.text import TextLanguage


class ConcatenativeSynthesis(TextToSignModel):
    """A class representing a Rule-Based model for translating text to sign language
    by concatenating sign language videos.

    Args:
        text_language (str | TextLanguage): (source) The text language or its identifier.
        sign_language (str | SignLanguage): (target) The sign language or its identifier.
        sign_features (str): (format) The sign features used for mapping labels to sign features.
    """

    def __init__(
        self,
        text_language: str | TextLanguage,
        sign_language: str | SignLanguage,
        sign_features: str,
    ) -> None:
        self._text_language = (
            get_text_language(text_language)
            if isinstance(text_language, str)
            else text_language
        )
        self._sign_language = (
            get_sign_language(sign_language)
            if isinstance(sign_language, str)
            else sign_language
        )
        self._sign_features = sign_features  # TODO: use feature loader

    @property
    def text_language(self):
        return self._text_language

    @property
    def sign_language(self):
        return self._sign_language

    @property
    def sign_features(self):
        return self._sign_features

    def translate(self, text: str):
        """
        Translate text to sign language.

        Args:
            text: The input text to be translated.

        Returns:
            The translated sign language sentence.

        """

        sign_language_sentence = None
        video_labels = []

        text = self.text_language.preprocess(text)
        sentences = self.text_language.sentence_tokenize(text)
        for sentence in sentences:
            tokens = self.text_language.tokenize(sentence)
            tags = self.text_language.get_tags(tokens)

            tokens, tags, contexts = self.sign_language.restructure_sentence(
                tokens, tags=tags
            )
            sign_dicts = self.sign_language.tokens_to_sign_dicts(
                tokens, tags=tags, contexts=contexts
            )

            video_labels.extend(
                [
                    label
                    for sign_dict in sign_dicts
                    for label in random.choices(
                        sign_dict[self.sign_language.SignDictKeys.SIGNS.value],
                        weights=sign_dict[self.sign_language.SignDictKeys.WEIGHTS.value],  # type: ignore
                        k=1,
                    )[0]
                ]
            )

        # TODO: map labels to sign features and concatenate
        sign_language_sentence = video_labels

        return sign_language_sentence



"""utility functions for models"""

from sign_language_translator.models.text_to_sign.concatenative_synthesis import (
    ConcatenativeSynthesis,
)


def get_model(
    model_code: str,
    sign_language: str | None = None,
    text_language: str | None = None,
    video_feature_model: str | None = None,
):
    if model_code.lower() in {
        "rule-based",
        "concatenative",
        "concatenativesynthesis",
        "concatenative-synthesis",
        "concatenative_synthesis",
    }:
        if text_language and sign_language and video_feature_model:
            # TODO: validate arg types
            return ConcatenativeSynthesis(
                text_language=text_language,
                sign_language=sign_language,
                sign_features=video_feature_model,
            )
        else:
            raise ValueError(
                "Inappropriate argument value for text_language, sign_language or video_feature_model"
            )
    elif model_code.lower() in ["stt", "sign-to-text", "sign_to_text"]:
        pass

    return None


__all__ = [
    "get_model",
]



"""other functions
"""

import re
from typing import Any, Iterable, List, Set, Tuple


def make_ngrams(sequence: Iterable, n: int) -> List[Iterable]:
    """Create all possible slices of the given iterable of size n.
    for example, sequence="1234" and n=2 would create ["12","23","34"].

    Args:
        sequence (Iterable): The iterable sequence from which the n-grams will be created.
        n (int): The size of the n-grams.

    Returns:
        List[Iterable]: A list of Iterables representing the n-grams created from the sequence.
            The type of list items is same as sequence argument.
    """

    start = 0
    end = len(sequence) - n  # type: ignore

    ngrams = (
        [sequence[i : i + n] for i in range(start, end + 1)] if end >= start else [] # type: ignore
    )
    return ngrams


def extract_supported_subsequences(
    sequence: Iterable[Any],
    tags: Iterable[Any],
    supported_tags: Set[Any],
    skipped_items: Set[Any],
) -> List[List[Any]]:
    """Extract supported subsequences from a sequence based on tags and skipped items.

    Args:
        sequence (Iterable[Any]): The input sequence.
        tags (Iterable[Any]): Tags corresponding to each item in the sequence.
        supported_tags (Set[Any]): Set of tags indicating support for a subsequence.
        skipped_items (Set[Any]): Set of items to be skipped.

    Returns:
        List[List[Any]]: A list of supported subsequences, where each inner list represents a subsequence.

    Examples:
        >>> sequence = [1, 2, 3, 4, 5, 6]
        >>> tags = ['A', 'A', 'B', 'A', 'A', 'C']
        >>> supported_tags = {'A'}
        >>> skipped_items = {2}
        >>> extract_supported_subsequences(sequence, tags, supported_tags, skipped_items)
        [[1], [4, 5]]
    """

    subsequences = []
    subseq = []
    for token, tag in zip(sequence, tags):
        if (tag in supported_tags) and (token not in skipped_items):
            subseq.append(token)
        else:
            if subseq:
                subsequences.append(subseq)
                subseq = []

    if subseq:
        subsequences.append(subseq)

    return subsequences


class ListRegex:
    """A utility class for finding sub-lists within a list of strings that match specified patterns.

    ListRegex provides methods for
    matching patterns against items in a list,
    searching for the first occurrence of patterns,
    finding all occurrences of patterns,
    and retrieving the starting and ending indices of matches.

    Patterns can be defined using:
        1. regular expressions (str)
        2. lists of patterns (regex (str) or a nested list of patterns)
        3. tuple containing the pattern and its interval quantifier ("\\w+", (2,None)).
    When using regular expressions, each pattern is matched against an individual item in the list.
    When using a list of patterns, any of the patterns in the list can match an item.
    When using a tuple of pattern and counts, items in the specified range can match the pattern.

    Example usage:

    ```python
    items = ["apple", "banana", "orange", "orange", "grape", "melon", "orange", "kiwi"]

    # Match the patterns against the items
    patterns = ["apple", "\\w+"]
    result = ListRegex.match(items, patterns)
    # Output: (0, 2)

    # Search for the first occurrence of the patterns
    patterns = [r"ba(na){2}", ("orange", (0,3))]
    result = ListRegex.search(items, patterns)
    # Output: (1, 4)

    # Find all occurrences of the patterns
    patterns = ["orange", ["grape", "kiwi"]]
    result = ListRegex.find_all(items, patterns)
    # Output: [['orange', 'grape'], ['orange', 'kiwi']]
    ```
    """

    @staticmethod
    def match(items: List[str], patterns: List[str | List | Tuple]) -> Tuple[int, int] | None:
        """
        Matches the given patterns against the items in the list.
        Applies the patterns at the start of the list of string.

        Args:
            items (List[str]): The sequence of strings to be matched.
            patterns (List[str|List]): The patterns to be matched against the items.

        Returns:
            Tuple[int, int] or None: A tuple containing the starting and ending indices of the matched items,
            or None if no match is found.
        """

        does_match = False
        pattern_index = 0
        item_index = 0
        count, counting_state = None, False

        while item_index < len(items) and pattern_index < len(patterns):
            pattern = patterns[pattern_index]
            item = items[item_index]

            # interval quantifiers: *, +, {n,m}
            # TODO: recursion on tuple
            if isinstance(pattern, tuple):
                assert (
                    len(pattern) == 2 and len(pattern[-1]) == 2
                ), "use proper syntax for repetition: (pattern, (min_count, max_count))"
                pattern, (min_count, max_count) = pattern[0], pattern[1]
                min_count = min_count if min_count is not None else float("-inf")
                max_count = max_count if max_count is not None else float("inf")
                if not counting_state:
                    count, counting_state = 0, True

            # Regex: abc, Character set: [abc],
            # TODO: Logical OR (ab|bc)  (sequence of items OR some other sequence of items)
            does_match = ListRegex._match_item(item, pattern)
            # TODO: shouldn't match when max_count is 0
            if does_match:
                item_index += 1
                if counting_state:
                    count += 1  # type: ignore
                if (count is not None) and count >= max_count:  # type: ignore
                    count, counting_state = None, False

                if not counting_state:
                    pattern_index += 1
            else:
                if counting_state:
                    pattern_index += 1
                    does_match = min_count <= count <= max_count  # type: ignore

                    count, counting_state = None, False

            if not does_match:
                break

        return (0, item_index) if does_match else None

    @staticmethod
    def _match_item(item: str, pattern):
        """
        Matches a single item against a pattern.
        A pattern could be a string regex or a list containing regex or list of regex and so on.

        Args:
            item (str): The item to be matched.
            pattern (str|list): The pattern to be matched against the item.

        Returns:
            bool: True if the item matches the pattern, False otherwise.
        """

        does_match = False
        if isinstance(pattern, str):
            does_match = bool(re.match(pattern, item))
        elif isinstance(pattern, list):
            does_match = any(ListRegex._match_item(item, pat) for pat in pattern)
        else:
            raise ValueError("unknown value of pattern, provide str or list")

        return does_match

    @staticmethod
    def search(items: List[str], patterns) -> Tuple[int, int] | None:
        """
        Searches for the first occurrence of the patterns in the list of items.

        Args:
            items (List[str]): The list of strings to be searched.
            patterns (List[str]): The patterns to be searched for in the items.

        Returns:
            Tuple[int, int] or None: A tuple containing the starting and ending indices of the matched items,
            or None if no match is found.
        """

        for i in range(len(items) - len(patterns) + 1):
            matched = ListRegex.match(items[i:], patterns)
            if matched:
                return (i, i + list(matched)[1])  # type: ignore
        return None

    @staticmethod
    def find_all(items: List[str], patterns: List) -> List[List[str]]:
        """Finds all occurrences of the patterns in the list of items.

        Args:
            items (List[str]): The list of strings to be searched.
            patterns (List[str]): The patterns to be searched for in the items.

        Returns:
            List[List[str]]: A list of matched subsequences of items.
        """

        matches = [
            items[start:end] for start, end in ListRegex.find_all_spans(items, patterns)
        ]
        return matches

    @staticmethod
    def find_all_spans(items: List[str], patterns: List) -> List[Tuple[int, int]]:
        """Finds the starting and ending indices of all occurrences of the patterns in the list of items.

        Args:
            items (List[str]): The list of strings to be searched.
            patterns (List[str]): The patterns to be searched for in the items.

        Returns:
            List[Tuple[int,int]]: A list of tuples containing the starting and ending indices of the matched items.
        """

        spans = []
        end = 0
        while end < len(items):
            span = ListRegex.search(items[end:], patterns)
            if span:
                span = end + span[0], end + span[1]
                _, end = span
                spans.append(span)
            else:
                break

        return spans


__all__ = [
    "make_ngrams",
    "extract_supported_subsequences",
    "ListRegex",
]



import enum
import re
from typing import Any, Callable, Iterable, List, Tuple


class Tags(enum.Enum):
    """Enumeration of token tags used in NLP processing."""

    WORD = "WORD"
    DATE = "DATE"
    TIME = "TIME"
    NAME = "NAME"
    SPACE = "SPACE"
    NUMBER = "NUMBER"
    ACRONYM = "ACRONYM"
    PUNCTUATION = "PUNCTUATION"
    SUPPORTED_WORD = "SUPPORTED_WORD"
    DEFAULT = ""
    END_OF_SEQUENCE = "EOS"
    START_OF_SEQUENCE = "SOS"
    # WORD_SENSE = "(anything)"
    # PERSON = {
    #     'id': 'PERSON',
    #     'name': 'Person',
    #     'description': 'Represents a person',
    #     'example': 'John Doe',
    # }

    def __repr__(self) -> str:
        return str(self)

    # :TODO: rename SUPPORTED to something better e.g. directly_mapped, recognized


class Rule:
    """A rule for token classification based on a matching function.

    Args:
        matcher (Callable[[str], bool]): A function that takes a token (str) as input
            and returns a boolean indicating whether the token matches the rule.
        tag (Any): The tag associated with tokens that match the rule.
        priority (int): The priority level of the rule.

    Methods:
        is_match(token: str) -> bool:
            Checks if the given token matches the rule.

        get_tag() -> str:
            Retrieves the tag associated with the rule.

        get_priority() -> int:
            Retrieves the priority level of the rule.

        from_pattern(pattern: str, tag: str, priority: int) -> Rule:
            Creates a rule from a regular expression pattern, tag, and priority.
            The created rule will use the pattern to match tokens.

    Note:
        - Rules with higher priority are applied first when classifying tokens.
        - The matcher function should return True if the token matches the rule, and False otherwise.
    """

    def __init__(self, matcher: Callable[[str], bool], tag: Any, priority: int):
        self.matcher = matcher
        self.tag = tag
        self.priority = priority

    def is_match(self, token: str):
        return self.matcher(token)

    def get_tag(self):
        return self.tag

    def get_priority(self):
        return self.priority

    @staticmethod
    def from_pattern(pattern: str, tag: Any, priority: int):
        def match(text: str):
            return bool(re.match(pattern, text))

        return Rule(match, tag, priority)


class Tagger:
    """A tagger that applies a set of rules to classify tokens.

    Args:
        rules (List[Rule]): A list of Rule objects representing the classification rules.
            Smaller priority value rules overwrite the others.
        default (Tags, optional): The default tag to assign when no rule matches a token.
            Defaults to Tags.DEFAULT.

    Methods:
        tag(tokens: List[str]) -> List[Tuple[str, Any]]:
            Assigns tags to a list of tokens based on the defined rules.
            Returns a list of tuples containing the token and its corresponding tag.

        get_tags(tokens: List[str]) -> List[Any]:
            Retrieves the tags for a list of tokens based on the defined rules.
            Returns a list of tags corresponding to the input tokens.

    Note:
        - The rules are applied in the order they appear in the list
            but higher priority (smaller value) rules overpower.
        - The default tag is assigned to tokens that do not match any rule.
    """

    def __init__(self, rules: List[Rule], default=Tags.DEFAULT):
        self.rules = rules
        self.default = default

    def tag(self, tokens: Iterable[str]) -> List[Tuple[str, Any]]:
        tagged_tokens = list(zip(tokens, self.get_tags(tokens)))

        return tagged_tokens

    def get_tags(self, tokens: Iterable[str]) -> List[Any]:
        return [self._apply_rules(token) for token in tokens]

    def _apply_rules(self, token: str) -> Any:
        tag = self.default
        priority = float("inf")

        for rule in self.rules:
            if rule.is_match(token):
                if rule.get_priority() < priority:
                    tag = rule.get_tag()
                    priority = rule.get_priority()

        return tag



import json
import os
import warnings
from collections import Counter
from glob import glob
from typing import Dict

from tqdm.auto import tqdm

from sign_language_translator.utils.tree import tree
from sign_language_translator import Settings


class DataCompletenessChecker:
    def __init__(
        self,
        signs_recordings_dir,
        signs_collection,
        sign_collection_parent_folder="videos",
        cameras=["front", "below", "left", "right"],
        persons=[
            f"person{n}"
            for n in [101, 151, 201, 202, 203, 204, 205, 251, 252, 253, 254, 255]
        ],
        sep=Settings.FILENAME_SEPARATOR,
        extension=".mp4",
        find_filepaths=False,
        load_filepaths=False,
    ) -> None:
        self.signs_recordings_dir = signs_recordings_dir
        self.sign_collection_parent_folder = sign_collection_parent_folder

        with open(os.path.join(signs_recordings_dir, "recordings_labels.json")) as f:
            self.RECORDINGS_LABELS = json.loads(f.read())

        if isinstance(signs_collection, str):
            self.signs_collection = signs_collection
            assert (
                self.signs_collection in self.RECORDINGS_LABELS
            ), "unsupported signs collections provided"
        else:
            NotImplementedError("provide str signs_collection")

        self.cameras = cameras
        self.persons = persons
        self.sep = sep
        self.extension = extension
        self.create_all_possible_filenames()

        self.paths_filename = (
            f"{self.sign_collection_parent_folder}_{self.signs_collection}_paths.txt"
        )

        self.filenames = set()
        self.file_paths = []

        if find_filepaths:
            self.find_filepaths()
        if load_filepaths:
            self.load_filepaths(append=True, load_all=False)

    def find_filepaths(self, append=False):
        patterns = [
            os.path.join(
                    self.signs_recordings_dir,
                    self.sign_collection_parent_folder,
                    self.signs_collection,
                    person,
                    f"*{camera}{self.extension}",
                )
            for person in self.persons
            for camera in self.cameras
        ]

        file_paths = [
            p
            for pattern in tqdm(patterns)
            for p in glob(pattern)
        ]
        file_paths = (
            list(set(file_paths) | set(self.file_paths))
            if append and hasattr(self, "file_paths")
            else file_paths
        )

        self.filenames = self.extract_filenames(file_paths, validate_repetition=True)
        self.file_paths = file_paths

    def load_filepaths(self, append=False, load_all=False):
        file_paths = []
        if os.path.exists(self.paths_filename):
            with open(self.paths_filename) as f:
                all_file_paths = f.read().split("\n")

            if load_all:
                file_paths = all_file_paths
            else:
                for p in all_file_paths:
                    dirname, fname = os.path.split(p)
                    fname = self.split_filename(fname)
                    if (
                        (fname[-1] in self.cameras)
                        and (fname[-2] in self.persons)
                        # and (fname[-2] == os.path.split(dirname)[-1].lower())
                    ):
                        file_paths.append(p)

        file_paths = (
            list(set(file_paths) | set(self.file_paths))
            if append and hasattr(self, "file_paths")
            else file_paths
        )

        self.filenames = self.extract_filenames(file_paths, validate_repetition=True)
        self.file_paths = file_paths

    def extract_filenames(self, file_paths, validate_repetition):
        filenames = [os.path.split(p)[-1] for p in file_paths]
        if validate_repetition:
            self.report_repetition(
                filenames, message="during extract_filenames(file_paths) - "
            )
        return set(filenames)

    def make_file_path(self, filename):
        filename_parts = self.split_filename(filename)
        return os.path.join(
            self.signs_recordings_dir,
            self.sign_collection_parent_folder,
            self.signs_collection,
            filename_parts[-2],  # person
            filename,
        )

    def report_repetition(self, data, message=""):
        counts = Counter(data)
        repeated = {k: v for k, v in counts.items() if v > 1}
        if len(repeated) > 0:
            raise Exception(
                f"{message}Repetition found for {len(repeated)} files:\n{repeated}"
            )

    def save_file_paths(self, overwrite=True):
        if not overwrite:
            self.load_filepaths(append=True, load_all=True)

        with open(self.paths_filename, "w") as f:
            f.write("\n".join(sorted(self.file_paths)))

    def create_all_possible_filenames(self):
        all_filenames = [
            f"{label}{self.sep}{person}{self.sep}{camera}{self.extension}"
            for label in self.RECORDINGS_LABELS[self.signs_collection]
            for person in self.persons
            for camera in self.cameras
        ]
        self.report_repetition(
            all_filenames, message="During create_all_possible_filenames() - "
        )

        self.all_filenames = set(all_filenames)

    def find_missing_files(self):
        return sorted(set(self.all_filenames) - set(self.filenames))

    def find_extra_files(self):
        return sorted(set(self.filenames) - set(self.all_filenames))

    def count_files(self, by: str, filenames=None):
        if filenames is None:
            filenames = self.filenames
        keys, idx = (
            (self.cameras, -1)
            if str(by).lower() in ["cam", "camera", "cameras"]
            else (self.persons, -2)
            if str(by).lower().startswith("person")
            else (self.RECORDINGS_LABELS[self.signs_collection], -3)
            if str(by).lower() in ["word", "words", "label", "labels"]
            else (None, None)
        )
        if keys is None:
            return dict(Counter(filenames))
        counts = {k: 0 for k in keys}
        counts.update(dict(Counter([self.split_filename(f)[idx] for f in filenames])))
        return counts

    def split_filename(self, fname):
        return fname[: -len(self.extension)].split(self.sep)

    def rename(
        self,
        label_replacements: Dict[str, str] = {},
        person_replacements: Dict[str, str] = {},
        camera_replacements: Dict[str, str] = {},
        path_filter=lambda path: True,
        n_files_donot_exist_examples=5,
        verbose=True,
        raise_or_log_FileExistsError="raise",
    ):
        n_renamed = 0
        files_donot_exist = []

        for i, fpath in enumerate(tqdm(self.file_paths, leave=False)):
            if not os.path.exists(fpath):
                files_donot_exist.append(fpath)
                continue
            if not path_filter(fpath):
                continue

            directory, fname = os.path.split(fpath)

            fname_parts = self.split_filename(fname)
            fname_parts[0] = label_replacements.get(fname_parts[0], fname_parts[0])
            fname_parts[1] = person_replacements.get(fname_parts[1], fname_parts[1])
            fname_parts[2] = camera_replacements.get(fname_parts[2], fname_parts[2])

            new_fname = self.sep.join(fname_parts) + self.extension
            new_fpath = os.path.join(directory, new_fname)

            if new_fname == fname:
                continue

            if os.path.exists(new_fpath):  # bloody os.rename on mac (T_T)
                if os.path.samefile(fpath, new_fpath):
                    warnings.warn(f'the filename strings are different ({fname= }, {new_fname= }) but they will effectively point to the same file, rename would likely skip this..')
                    input_txt = ''
                    while input_txt not in ['yes', 'no','y', 'n']:
                        input_txt = input('continue? (y/n)')
                    if input_txt.lower() in ['no', 'n']:
                        break
                else:
                    FileExistsError_message = f"{new_fpath} named file already exists"
                    if raise_or_log_FileExistsError == "log":
                        print(FileExistsError_message)
                        continue
                    else:
                        raise FileExistsError(FileExistsError_message)

            if new_fname in self.filenames:
                FileExistsError_message = (
                    f"{new_fname} named file already exists somewhere in the collection"
                )
                if raise_or_log_FileExistsError == "log":
                    print(FileExistsError_message)
                    continue
                else:
                    raise FileExistsError(FileExistsError_message)

            else:
                os.rename(fpath, new_fpath)
                self.file_paths[i] = new_fpath
                self.filenames.remove(fname)
                self.filenames.add(new_fname)
                n_renamed += 1

        if verbose:
            print(
                f"{n_renamed} files renamed out of total {len(self.file_paths)} files scanned."
            )

        if len(files_donot_exist) > 0:
            examples = "\n\t" + "\n\t".join(
                files_donot_exist[:n_files_donot_exist_examples]
            )
            warnings.warn(
                f"{len(files_donot_exist)} files do not exist. Here are a few examples:{examples}"
            )

    def make_tree(
        self,
        n_levels_above_signs_collection: int = 0,
        directory_only=True,
        extra_line=True,
        ignore=["__pycache__", "temp"],
    ):
        root_dir = os.path.join(
            self.signs_recordings_dir,
            self.sign_collection_parent_folder,
            self.signs_collection,
        )
        for _ in range(n_levels_above_signs_collection):
            root_dir = os.path.split(root_dir)[0]

        tree(
            root_dir,
            directory_only=directory_only,
            extra_line=extra_line,
            ignore=ignore,
        )



"""generate subtitles in WebVTT format
"""

from typing import List, Tuple


def _make_timestamp(total_seconds) -> str:
    hours, remaining_seconds = divmod(total_seconds, 3600)
    minutes, seconds = divmod(remaining_seconds, 60)

    return f"{int(hours):02d}:{int(minutes):02d}:{seconds:06.3f}"


def make_captions(
    texts: List[str],
    durations: List[float],
    highlighted_ranges: List[Tuple[int, int]],
    vertical_position_frac: float = 0.6,
    align: str = "left",
    emphasis_tag="b",
) -> str:

    captions = ["WEBVTT"]

    elapsed_time = 0.0
    for text, duration, (h_start, h_end) in zip(texts, durations, highlighted_ranges):
        captions.append("")
        captions.append(
            f"{_make_timestamp(elapsed_time)} --> {_make_timestamp(elapsed_time + duration)}"
            + f" line:{vertical_position_frac:.0%}"
            + f" align:{align}"
        )
        captions.append(
            text[:h_start]
            + f"<{emphasis_tag}>"
            + text[h_start:h_end]
            + f"</{emphasis_tag}>"
            + text[h_end:]
        )
        elapsed_time += duration

    return "\n".join(captions)


def make_chapters(
    texts: List[str], durations: List[float], highlighted_ranges: List[Tuple[int, int]]
) -> str:
    return make_captions(texts, durations, highlighted_ranges)



"""encapsulates landmarks specific information such as how to connect them and color those connections
"""

from typing import List, Tuple, Union

import numpy as np
import torch


class LandmarksInfo:
    """encapsulates the ways to connect landmarks and color those connections"""

    POSE_CONNECTIONS = [
        [0, 1],
        [0, 4],
        [1, 2],
        [2, 3],
        [3, 7],
        [4, 5],
        [5, 6],
        [6, 8],
        [9, 10],
        [11, 12],
        [11, 13],
        [11, 23],
        [12, 14],
        [12, 24],
        [13, 15],
        [14, 16],
        [15, 17],
        [15, 19],
        [15, 21],
        [16, 18],
        [16, 20],
        [16, 22],
        [17, 19],
        [18, 20],
        [23, 24],
        [23, 25],
        [24, 26],
        [25, 27],
        [26, 28],
        [27, 29],
        [27, 31],
        [28, 30],
        [28, 32],
        [29, 31],
        [30, 32],
    ]
    HAND_CONNECTIONS = [
        [0, 1],
        [0, 5],
        [0, 17],
        [1, 2],
        [2, 3],
        [3, 4],
        [5, 6],
        [5, 9],
        [6, 7],
        [7, 8],
        [9, 10],
        [9, 13],
        [10, 11],
        [11, 12],
        [13, 14],
        [13, 17],
        [14, 15],
        [15, 16],
        [17, 18],
        [18, 19],
        [19, 20],
    ]

    N_POSE_LANDMARKS = 33
    N_HAND_LANDMARKS = 21

    POSE_LEFT_WRIST_INDEX = 15
    POSE_RIGHT_WRIST_INDEX = 16
    HAND_WRIST_INDEX = 0

    BOTH_HANDS_CONNECTIONS = (
        sorted(HAND_CONNECTIONS)
        + (np.array(sorted(HAND_CONNECTIONS)) + N_HAND_LANDMARKS).tolist()
    )
    ALL_CONNECTIONS = (
        sorted(POSE_CONNECTIONS)
        + (np.array(BOTH_HANDS_CONNECTIONS) + N_POSE_LANDMARKS).tolist()
    )

    N_BOTH_HANDS_LANDMARKS = 2 * N_HAND_LANDMARKS
    N_ALL_LANDMARKS = N_POSE_LANDMARKS + 2 * N_HAND_LANDMARKS

    N_POSE_FEATURES = N_POSE_LANDMARKS * 4
    N_HAND_FEATURES = N_HAND_LANDMARKS * 3
    N_BOTH_HANDS_FEATURES = N_HAND_FEATURES * 2
    N_ALL_FEATURES = N_POSE_FEATURES + N_BOTH_HANDS_FEATURES

    VISIBILITY_INDEXES = (np.arange(N_POSE_LANDMARKS) * 4 + 3).tolist()

    def infer_connections(n_landmarks: int) -> List[List[int]]:
        """find an appropriate connections list based on number of landmarks

        Args:
            n_landmarks (int): number of landmarks

        Returns:
            List[List[int]]: Pairs of indices of the landmarks array to be connected by a line. Defaults to [].
        """

        return (
            LandmarksInfo.POSE_CONNECTIONS
            if n_landmarks == LandmarksInfo.N_POSE_LANDMARKS
            else LandmarksInfo.HAND_CONNECTIONS
            if n_landmarks == LandmarksInfo.N_HAND_LANDMARKS
            else LandmarksInfo.BOTH_HANDS_CONNECTIONS
            if n_landmarks == LandmarksInfo.N_BOTH_HANDS_LANDMARKS
            else LandmarksInfo.ALL_CONNECTIONS
            if n_landmarks == LandmarksInfo.N_ALL_LANDMARKS
            else []
        )

        # def repeat(lt, n_final):
        #     num_repeats = int(n_final / len(lt))

        #     if num_repeats == 1:
        #         return lt

        #     arr = np.array(lt)
        #     return np.concatenate([arr + i for i in range(num_repeats)]).tolist()

        # print(n_landmarks)
        # return (
        #     repeat(LandmarksInfo.POSE_CONNECTIONS, n_landmarks)
        #     if n_landmarks % LandmarksInfo.N_POSE_LANDMARKS == 0
        #     else repeat(LandmarksInfo.HAND_CONNECTIONS, n_landmarks)
        #     if n_landmarks % LandmarksInfo.N_HAND_LANDMARKS == 0
        #     else repeat(LandmarksInfo.ALL_CONNECTIONS, n_landmarks)
        #     if n_landmarks % LandmarksInfo.N_ALL_LANDMARKS == 0
        #     else []
        # )

    class Colors:
        """encapsulates colors applied to landmarks"""

        BLACK = (0, 0, 0)
        RED = (1, 0, 0)
        GREEN = (0, 1, 0)
        BLUE = (0, 0, 1)
        CYAN = (0, 1, 1)
        MAGENTA = (1, 0, 1)

        DEFAULT_CONNECTION = BLACK
        DEFAULT_LANDMARK = RED
        LEFT = CYAN
        RIGHT = MAGENTA
        THUMB = GREEN
        PINKY = BLUE

        def _get_default_connection_color(
            start: int = None, end: int = None
        ) -> Tuple[float, float, float]:
            """get a color based on which part of body is being drawn

            Args:
                start (int, optional): landmark index where line starts. Defaults to None.
                end (int, optional): landmark index where line ends. Defaults to None.

            Returns:
                Tuple[float, float, float]: RGB color value [0 to 1]
            """

            return LandmarksInfo.Colors.DEFAULT_CONNECTION

        def _get_pose_connection_color(
            start: int, end: int
        ) -> Tuple[float, float, float]:
            """get a color based on which part of body is being drawn

            Args:
                start (int, optional): landmark index where line starts. Defaults to None.
                end (int, optional): landmark index where line ends. Defaults to None.

            Returns:
                Tuple[float, float, float]: RGB color value [0 to 1]
            """

            return (
                LandmarksInfo.Colors.LEFT
                if (start % 2 == 1 and end % 2 == 1)
                else LandmarksInfo.Colors.RIGHT
                if (start % 2 == 0 and end % 2 == 0)
                else LandmarksInfo.Colors._get_default_connection_color(start, end)
            )

        def _get_hand_connection_color(
            start: int, end: int
        ) -> Tuple[float, float, float]:
            """get a color based on which part of body is being drawn

            Args:
                start (int, optional): landmark index where line starts. Defaults to None.
                end (int, optional): landmark index where line ends. Defaults to None.

            Returns:
                Tuple[float, float, float]: RGB color value [0 to 1]
            """

            return (
                LandmarksInfo.Colors.THUMB
                if end <= 4
                else LandmarksInfo.Colors.PINKY
                if start >= 17
                else LandmarksInfo.Colors._get_default_connection_color(start, end)
            )

        def _get_both_hands_connection_color(
            start: int, end: int
        ) -> Tuple[float, float, float]:
            """get a color based on which part of body is being drawn

            Args:
                start (int, optional): landmark index where line starts. Defaults to None.
                end (int, optional): landmark index where line ends. Defaults to None.

            Returns:
                Tuple[float, float, float]: RGB color value [0 to 1]
            """

            return (
                LandmarksInfo.Colors._get_hand_connection_color(start, end)
                if start < LandmarksInfo.N_HAND_LANDMARKS
                and end < LandmarksInfo.N_HAND_LANDMARKS
                else LandmarksInfo.Colors._get_hand_connection_color(
                    start - LandmarksInfo.N_HAND_LANDMARKS,
                    end - LandmarksInfo.N_HAND_LANDMARKS,
                )
            )

        def _get_all_connection_color(
            start: int, end: int
        ) -> Tuple[float, float, float]:
            """get a color based on which part of body is being drawn

            Args:
                start (int, optional): landmark index where line starts. Defaults to None.
                end (int, optional): landmark index where line ends. Defaults to None.

            Returns:
                Tuple[float, float, float]: RGB color value [0 to 1]
            """

            if (
                start < LandmarksInfo.N_POSE_LANDMARKS
                and end < LandmarksInfo.N_POSE_LANDMARKS
            ):
                color = LandmarksInfo.Colors._get_pose_connection_color(start, end)

            elif (
                start < LandmarksInfo.N_ALL_LANDMARKS
                and end < LandmarksInfo.N_ALL_LANDMARKS
            ):
                color = LandmarksInfo.Colors._get_both_hands_connection_color(
                    start - LandmarksInfo.N_POSE_LANDMARKS,
                    end - LandmarksInfo.N_POSE_LANDMARKS,
                )

            else:
                color = LandmarksInfo.Colors._get_default_connection_color(start, end)

            return color

    class Reshaper:
        """[USE THE SAME OBJECT TO FOLD & UNFOLD and only ONE array at a time.]
        Convert arrays into (..., n_landmarks, (x,y,z)) and back (..., n_landmarks*(3|4)).
        The problem is created by the visibility feature (x,y,z, v )
        which is available for pose only so stacking in 3D becomes jagged."""

        def __init__(self) -> None:
            """__init__ function. initializes variables that store infromation removed during folding with None."""

            self.last_dim_size = None
            self.landmarks_visibility_backup = None

        def fold(
            self, landmarks: Union[np.ndarray, torch.Tensor]
        ) -> Union[np.ndarray, torch.Tensor]:
            """Turn a (..., m*n+k) array into (..., m, n). The cut out k values are stored within this object.

            Args:
                landmarks (Union[np.ndarray, torch.Tensor]): the (x,y,z,[v]) body and hands landmarks of a body.

            Raises:
                NotImplementedError: use numpy.ndarray or torch.Tensor landmarks

            Returns:
                Union[np.ndarray, torch.Tensor]: the folded (x,y,z) body and hands landmarks of a body.
            """

            if isinstance(landmarks, torch.Tensor):
                concatenate = torch.concatenate
            elif isinstance(landmarks, np.ndarray):
                concatenate = np.concatenate
            else:
                raise NotImplementedError("use numpy.ndarray or torch.Tensor")

            self.last_dim_size = landmarks.shape[-1]

            if self.last_dim_size == 4:
                self.landmarks_visibility_backup = landmarks[..., 3:]#.copy()
                return landmarks[..., :3]

            elif self.last_dim_size == LandmarksInfo.N_POSE_FEATURES:
                landmarks = landmarks.reshape(landmarks.shape[:-1] + (-1, 4))
                self.landmarks_visibility_backup = landmarks[..., 3:]#.copy()
                return landmarks[..., :3]

            elif self.last_dim_size == LandmarksInfo.N_ALL_FEATURES:
                pose = landmarks[..., : LandmarksInfo.N_POSE_FEATURES]
                rest = landmarks[..., LandmarksInfo.N_POSE_FEATURES :]

                pose = pose.reshape(pose.shape[:-1] + (-1, 4))
                rest = rest.reshape(rest.shape[:-1] + (-1, 3))

                self.landmarks_visibility_backup = pose[..., 3:]#.copy()
                return concatenate([pose[..., :3], rest], axis=-2)

            elif self.last_dim_size in [
                LandmarksInfo.N_HAND_FEATURES,
                LandmarksInfo.N_BOTH_HANDS_FEATURES,
            ]:
                landmarks = landmarks.reshape(landmarks.shape[:-1] + (-1, 3))

            self.landmarks_visibility_backup = None

            return landmarks

        def unfold(
            self, landmarks: Union[np.ndarray, torch.Tensor]
        ) -> Union[np.ndarray, torch.Tensor]:
            """turn a (..., m, n) array into (..., m*n+k). The k values are pasted from values stored within this object during folding.

            Args:
                landmarks (Union[np.ndarray, torch.Tensor]): the folded (x,y,z) body and hands landmarks of a body.

            Raises:
                NotImplementedError: use numpy.ndarray or torch.Tensor landmarks

            Returns:
                Union[np.ndarray, torch.Tensor]: the unfolded (x,y,z,[v]) body and hands landmarks of a body.
            """

            if isinstance(landmarks, torch.Tensor):
                concatenate = torch.concatenate
            elif isinstance(landmarks, np.ndarray):
                concatenate = np.concatenate
            else:
                raise NotImplementedError("use numpy.ndarray or torch.Tensor")

            if self.last_dim_size == 4:
                unfolded = concatenate(
                    [landmarks, self.landmarks_visibility_backup], axis=-1
                )

            elif self.last_dim_size == LandmarksInfo.N_POSE_FEATURES:
                unfolded = concatenate(
                    [landmarks, self.landmarks_visibility_backup], axis=-1
                )
                unfolded = unfolded.reshape(landmarks.shape[:-2] + (-1,))

            elif self.last_dim_size == LandmarksInfo.N_ALL_FEATURES:
                pose = concatenate(
                    [
                        landmarks[..., : LandmarksInfo.N_POSE_LANDMARKS, :],
                        self.landmarks_visibility_backup,
                    ],
                    axis=-1,
                )
                rest = landmarks[..., LandmarksInfo.N_POSE_LANDMARKS :, :]

                pose = pose.reshape(pose.shape[:-2] + (-1,))
                rest = rest.reshape(rest.shape[:-2] + (-1,))

                unfolded = concatenate([pose, rest], axis=-1)

            elif self.last_dim_size in [
                LandmarksInfo.N_HAND_FEATURES,
                LandmarksInfo.N_BOTH_HANDS_FEATURES,
            ]:
                unfolded = landmarks.reshape(landmarks.shape[:-2] + (-1,))

            else:
                unfolded = landmarks

            self.last_dim_size = None
            self.landmarks_visibility_backup = None

            return unfolded



"""concatenate multiple clips (video & pose) with transitions

Raises:
    NotImplementedError: numpy.ndarray or torch.Tensor
    ValueError: various
"""

from typing import List, Tuple, Union

import moviepy.editor as mpy
import numpy as np
import torch
from scipy.signal import find_peaks

from ..utils.landmarks_info import LandmarksInfo
from . import transforms as T


def _find_edge_peaks_heights(
    values: Union[List[float], np.ndarray, torch.Tensor],
    tolarance_val: float = 0.02,
    height_frac: float = 0.7,
    max_n_steps: int = 25,
) -> Tuple[float, float]:
    """finds the heights of the peaks near the edges of the data

    Args:
        values (Union[List[float], np.ndarray, torch.Tensor]): the timeseries data or y_values of a function e.g. vertical position of a wrist over time
        tolarance_val (float, optional): how much rise from initial height should be ignored. this plus the minimum value in data defines a threshold. Peaks below the threshold are discarded. Defaults to 0.02.
        height_frac (float, optional): the fraction of the edge peak height relative to the lowest level to return. Defaults to 0.7.
        max_n_steps (int, optional): maximum number of steps from each side of the data to consider for a peak. Defaults to 25.

    Raises:
        NotImplementedError: provide numpy.ndarray or torch.Tensor values

    Returns:
        Tuple[int,int]: pair of heights of edge peaks
    """

    values = np.array(values)
    num_values = len(values)

    left_lowest = min(values[:max_n_steps])
    right_lowest = min(values[-max_n_steps:])

    peaks_indexes = find_peaks(values)[0]
    peaks_indexes = np.sort(peaks_indexes)

    left_peaks_indexes = peaks_indexes[peaks_indexes < max_n_steps]
    right_peaks_indexes = peaks_indexes[peaks_indexes >= num_values - max_n_steps]

    valid_left_peaks_indexes = left_peaks_indexes[
        values[left_peaks_indexes] > (left_lowest + tolarance_val)
    ]
    valid_right_peaks_indexes = right_peaks_indexes[
        values[right_peaks_indexes] > (right_lowest + tolarance_val)
    ]

    left_height = (
        values[valid_left_peaks_indexes[0]]
        if len(valid_left_peaks_indexes) > 0
        else max(values[:max_n_steps])
    )
    right_height = (
        values[valid_right_peaks_indexes[-1]]
        if len(valid_left_peaks_indexes) > 0
        else max(values[-max_n_steps:])
    )

    return (
        left_height * height_frac + (1 - height_frac) * left_lowest,
        right_height * height_frac + (1 - height_frac) * right_lowest,
    )


def find_cut_indexes(
    landmarks_list: List[Union[np.ndarray, torch.Tensor]],
    coordinate_type: str = "2D",
    landmark_number_to_track: Union[int, List[int]] = 16,
    landmarks_to_use_for_rectification: List[int] = [11, 12, 23, 24],
    max_cut_steps: int = 25,
    tolarance_val: float = 0.02,
    height_frac: float = 0.7,
    rectify: bool = True,
) -> List[Tuple[int, int]]:
    """find the index numbers near to edges where a landmark clip should be cut. these are the points where a tracked landmark is at a prticular height.

    Args:
        landmarks_list (List[Union[np.ndarray, torch.Tensor]]): list of multi_frame_landmarks which are 2D or 3D sequences of body landmarks.
        coordinate_type (str, optional): the type of coodinates in the landmarks. If the are image coordinates, use "2D". if they are world coordinates, use "3D". Defaults to "2D".
        landmark_number_to_track (Union[int, List[int]], optional): movement of which landmark should be considered to find cut indexes. Defaults to 16 (right wrist).
        landmarks_to_use_for_rectification (List[int], optional): the index of landmarks that should be used stabilization. Defaults to (torso) [11, 12, 23, 24].
        max_cut_steps (int, optional): how many indexes to cut at most from one side. Defaults to 25.
        tolarance_val (float, optional): how much rise from initial height should be ignored. Defaults to 0.02.
        height_frac (float, optional): what fraction of landmark's max or peak height should be used to cut at. Defaults to 0.7.
        rectify (bool, optional): stabilize landmark_list into upright position before tracking the landmark. Defaults to True.

    Raises:
        ValueError: use "2d" or "3d" as coordinate_type
        ValueError: use int or list of int as landmark_number_to_track

    Returns:
        List[Tuple[int, int]]: list of pairs of indexes where multi_frame_landmarks should be cut.
    """

    reshapers = [LandmarksInfo.Reshaper() for _ in range(len(landmarks_list))]
    landmarks_list = [
        reshaper.fold(landmarks)
        for landmarks, reshaper in zip(landmarks_list, reshapers)
    ]

    if str(coordinate_type).lower() == "3d":
        if rectify:
            landmarks_list = T.stabilize_landmarks(
                landmarks_list,
                landmarks_to_use=landmarks_to_use_for_rectification,
                coordinates_to_use=[0, 1, 2],
                target=[
                    [0.16, -0.5, 0],
                    [-0.16, -0.5, 0],
                    [0.10, 0, 0],
                    [-0.10, 0, 0],
                ],
                infer_target=False,
            )

    elif str(coordinate_type).lower() == "2d":
        if rectify:
            landmarks_list = T.stabilize_landmarks(
                landmarks_list,
                landmarks_to_use=landmarks_to_use_for_rectification,
                coordinates_to_use=[0, 1],
                infer_target=True,
            )

    else:
        raise ValueError('use "2d" or "3d" as coordinate_type')

    if isinstance(landmark_number_to_track, int):
        landmark_number_to_track = [landmark_number_to_track] * len(landmarks_list)

    elif isinstance(landmark_number_to_track, list):
        assert len(landmark_number_to_track) == len(
            landmarks_list
        ), "please specify as many landmarks to track as items in landmark_list"
        assert all(
            [isinstance(num, int) for num in landmark_number_to_track]
        ), "please use int landmark_number_to_track"

    else:
        raise ValueError("use int or list of int as landmark_number_to_track")

    peaks_heights = [
        _find_edge_peaks_heights(
            -landmarks[:, landmark_num_to_track, 1],
            tolarance_val=tolarance_val,
            height_frac=height_frac,
            max_n_steps=max_cut_steps,
        )
        for landmarks, landmark_num_to_track in zip(
            landmarks_list, landmark_number_to_track
        )
    ]

    cut_indexes = [[0, len(landmarks) - 1] for landmarks in landmarks_list]
    for i in range(len(landmarks_list) - 1):

        cut_height = min(peaks_heights[i][1], peaks_heights[i + 1][0])

        mask = -landmarks_list[i][:, landmark_number_to_track[i], 1] >= cut_height
        cut_indexes[i][1] = len(landmarks_list[i]) - 1 - np.array(mask)[::-1].argmax()

        mask = (
            -landmarks_list[i + 1][:, landmark_number_to_track[i + 1], 1] >= cut_height
        )
        cut_indexes[i + 1][0] = np.array(mask).argmax()

    return cut_indexes


def insert_transitions(
    landmarks_list: List[Union[np.ndarray, torch.Tensor]], n_transition_frames: int
) -> List[Union[np.ndarray, torch.Tensor]]:
    """calculate transitional frames from edge frames of two landmark_clips using linear interpolation and insert it between them.

    Args:
        landmarks_list (List[np.ndarray | torch.Tensor]): list of multi_frame_landmarks which are 2D or 3D sequences of body landmarks.
        n_transition_frames (int): number of transtional frames to calculate and insert between two landmark_clips

    Raises:
        NotImplementedError: use numpy.ndarray or torch.Tensor landmarks

    Returns:
        List[np.ndarray | torch.Tensor]: same as input list but with transition clips inserted.
    """

    if len(landmarks_list) <= 1 or n_transition_frames == 0:
        return landmarks_list

    if isinstance(landmarks_list[0], np.ndarray):
        concatenate = np.concatenate
    elif isinstance(landmarks_list[0], torch.Tensor):
        concatenate = torch.concatenate
    else:
        raise NotImplementedError("use numpy.ndarray or torch.Tensor")

    # make transitions
    edge_frames = concatenate(
        [
            frame
            for (landmarks_1, landmarks_2) in zip(
                landmarks_list[:-1], landmarks_list[1:]
            )
            for frame in (landmarks_1[-1:], landmarks_2[:1])
        ]
    )
    n_frames = len(edge_frames)
    old_timesteps = np.arange(n_frames)

    # correct timesteps to create transitions only:
    # n_f=4, n_t=1 --> [0, 0.5, 1, 1.5, 2, 2.5, 3] --> [0.5, 2.5]
    new_timesteps = (
        np.linspace(0, n_frames - 1, n_transition_frames * (n_frames - 1) + n_frames)[
            :-1
        ]
        .reshape((-1, n_transition_frames + 1))[0::2, 1:]
        .reshape(-1)
    )

    transition_frames = T._interpolate_multi_frame_landmarks(
        edge_frames,
        old_timesteps,
        new_timesteps,
        time_dimention=0,
    ).reshape((-1, n_transition_frames) + edge_frames[0].shape)

    # insert transitions
    landmarks_list = [
        item
        for landmarks, transition in zip(landmarks_list[:-1], transition_frames)
        for item in (landmarks, transition)
    ] + landmarks_list[-1:]

    return landmarks_list


def concatenate_landmarks(
    landmarks_list: List[Union[np.ndarray, torch.Tensor]],
    n_transition_frames: int = 0,
    trim: bool = False,
    cut_indexes: List[Tuple[int, int]] = None,
    infer_cut_indexes: bool = True,
    landmark_number_to_track: Union[int, List[int]] = 16,
    coordinate_type: str = "3D",
    max_cut_steps: int = 25,
    tolarance_val: float = 0.02,
    height_frac: float = 0.7,
    rectify_for_infering_cut_indexes: bool = True,
    landmarks_to_use_for_rectification: List[int] = [11, 12, 23, 24],
) -> Union[np.ndarray, torch.Tensor]:
    """Concatenate a list of multi_frame_landmarks into one. optionally trim the clips to form continuous sentences without having the hand go back to resting position after each word. Also optionally insert transition frames between clips to avoid "choppy" cuts.

    Args:
        landmarks_list (List[np.ndarray | torch.Tensor]): list of multi_frame_landmarks which are 2D or 3D sequences of body landmarks.
        n_transition_frames (int, optional): number of transtional frames to calculate and insert between two landmark_clips. Defaults to 0.
        trim (bool, optional): whether to cut out parts between the landmark_clips where the hand is just rising from or going down to the resting position. Defaults to False.
        cut_indexes (List[Tuple[int, int]], optional): list pairs of indexes where multi_frame_landmarks should be cut. Defaults to None.
        infer_cut_indexes (bool, optional): whether to use provided cut_index_list (False) or figure them out (True). Defaults to True.
        landmark_number_to_track (Union[int, List[int]], optional): movement of which landmark should be considered to find cut indexes. Defaults to 16 (right wrist).
        coordinate_type (str, optional): the type of coodinates in the landmarks. If the are image coordinates, use "2D". if they are world coordinates, use "3D". Defaults to "3D".
        max_cut_steps (int, optional): how many indexes to cut at most from one side. Defaults to 25.
        tolarance_val (float, optional): how much rise from initial height should be ignored (for triming). Defaults to 0.02.
        height_frac (float, optional): what fraction of landmark's max or peak height should be used to cut at. Defaults to 0.7.
        rectify_for_infering_cut_indexes (bool, optional): stabilize landmark_list into upright position before tracking the landmark. Defaults to True.
        landmarks_to_use_for_rectification (List[int], optional): the index of landmarks that should be used stabilization. Defaults to (torso) [11, 12, 23, 24].

    Raises:
        ValueError: Empty landmark_list. Nothing to concatenate.
        ValueError: must provide cut_indexes or set infer_cut_indexes True
        NotImplementedError: use numpy.ndarray or torch.Tensor

    Returns:
        Union[np.ndarray, torch.Tensor]: a single sequence of landmarks
    """

    if len(landmarks_list) == 0:
        raise ValueError("Empty landmark_list. Nothing to concatenate.")

    if trim:
        if cut_indexes is None:
            if infer_cut_indexes:
                cut_indexes = find_cut_indexes(
                    landmarks_list,
                    coordinate_type=coordinate_type,
                    landmark_number_to_track=landmark_number_to_track,
                    rectify=rectify_for_infering_cut_indexes,
                    landmarks_to_use_for_rectification=landmarks_to_use_for_rectification,
                    max_cut_steps=max_cut_steps,
                    tolarance_val=tolarance_val,
                    height_frac=height_frac,
                )
            else:
                raise ValueError(
                    "must provide cut_indexes or set infer_cut_indexes True"
                )

        landmarks_list = [
            landmarks[start : end + 1]
            for landmarks, (start, end) in zip(landmarks_list, cut_indexes)
        ]

    if n_transition_frames > 0:
        # round(np.random.normal(4,1,(1,)).clip(0))  # or .clip(2,6) or .clip(2)
        landmarks_list = insert_transitions(landmarks_list, n_transition_frames)

    if isinstance(landmarks_list[0], np.ndarray):
        concatenate = np.concatenate
    elif isinstance(landmarks_list[0], torch.Tensor):
        concatenate = torch.concatenate
    else:
        raise NotImplementedError("use numpy.ndarray or torch.Tensor")

    return concatenate(landmarks_list)


def concatenate_clips(
    clips: List[mpy.VideoFileClip],
    n_transition_frames: int = 0,
    trim: bool = False,
    return_trimmed_only:bool=False,
    triming_reference_landmarks: List[Union[np.ndarray, torch.Tensor]] = None,
    cut_indexes: List[Tuple[int, int]] = None,
    coordinate_type: str = "2D",
    landmark_number_to_track: Union[int, List[int]] = 16,
    landmarks_to_use_for_rectification: List[int] = [11, 12, 23, 24],
    max_cut_steps: int = 25,
    tolarance_val: float = 0.02,
    height_frac: float = 0.7,
    rectify_landmarks: bool = False,
) -> mpy.VideoFileClip:
    """Concatenate a list of video clips into one. optionally trim the clips to form continuous sentences without having the hand go back to resting position after each word. optionally also insert transition frames between clips to avoid "choppy" cuts.

    Args:
        clips (List[mpy.VideoFileClip]): list of video clips
        n_transition_frames (int, optional): number of transtional frames to calculate and insert between two landmark_clips. Defaults to 0.
        trim (bool, optional): whether to cut out parts between the clips where the hand is just rising from or going down to the resting position. Defaults to False.
        triming_reference_landmarks (List[np.ndarray | torch.Tensor], optional): the landmarks extracted from the clips. Defaults to None.
        cut_indexes (List[Tuple[int, int]], optional): list pairs of indexes where the provided multi_frame_landmarks should be cut. Defaults to None.
        coordinate_type (str, optional): the type of coodinates in the landmarks. If they are image coordinates, use "2D". if they are world coordinates, use "3D". Defaults to "2D".
        landmark_number_to_track (Union[int, List[int]], optional): movement of which landmark should be considered to find cut indexes. Defaults to 16 (right wrist).
        landmarks_to_use_for_rectification (List[int], optional): the index of landmarks that should be used stabilization. Defaults to (torso) [11, 12, 23, 24].
        max_cut_steps (int, optional): how many frames to cut at most from one side. Defaults to 25.
        tolarance_val (float, optional): how much rise from initial height should be ignored (for triming). Defaults to 0.02.
        height_frac (float, optional): what fraction of landmark's max or peak height should be used to cut at. Defaults to 0.7.
        rectify_landmarks (bool, optional): stabilize landmark_list into upright position before tracking the landmark. Defaults to False.

    Raises:
        ValueError: provide triming_reference_landmarks or cut_indexes

    Returns:
        mpy.VideoFileClip: a single video clip
    """

    if trim:
        if cut_indexes is None:
            if triming_reference_landmarks is None:
                raise ValueError("provide triming_reference_landmarks or cut_indexes")

            cut_indexes = find_cut_indexes(
                triming_reference_landmarks,
                coordinate_type=coordinate_type,
                landmark_number_to_track=landmark_number_to_track,
                landmarks_to_use_for_rectification=landmarks_to_use_for_rectification,
                max_cut_steps=max_cut_steps,
                tolarance_val=tolarance_val,
                height_frac=height_frac,
                rectify=rectify_landmarks,
            )

        cut_durations = [
            [
                clip.duration * cut_index[0] / (len(landmarks) - 1),
                clip.duration * cut_index[1] / (len(landmarks) - 1),
            ]
            for clip, cut_index, landmarks in zip(
                clips, cut_indexes, triming_reference_landmarks
            )
        ]

        clips = [
            clip.subclip(cut_duration[0], cut_duration[1])
            for clip, cut_duration in zip(clips, cut_durations)
        ]

    if return_trimmed_only:
        return clips

    if n_transition_frames > 0:
        final = mpy.concatenate_videoclips(
            clips[:1]
            + [clip.crossfadein(n_transition_frames / clip.fps) for clip in clips[1:]],
            padding=-n_transition_frames / clips[0].fps,
            method="compose",
        )
    else:
        final = mpy.concatenate_videoclips(clips, method="compose")

    return final



"""This module contains various helper functions and utilities for configuration purposes.

Module Structure:
- set_dataset_dir(path: str) -> None: Sets the dataset directory path.
"""

import os
from .settings import Settings


def set_dataset_dir(path: str) -> None:
    """Set the sign-language-datasets directory path.

    Args:
        path (str): The path to the dataset directory.

    Raises:
        AssertionError: If the provided path is not a directory.
    """

    assert os.path.isdir(path), "the provided path is not a directory"

    Settings.DATASET_ROOT_DIRECTORY = path
    # ? trigger an event



from abc import ABC, abstractmethod, abstractproperty
from typing import Any, Union, Iterable


class TextToSignModel(ABC):
    @abstractproperty
    def sign_language(self):
        pass

    @abstractproperty
    def text_language(self):
        pass

    @abstractproperty
    def sign_features(self):
        pass

    @abstractmethod
    def translate(self, text: Union[str, Iterable[str]]) -> Any: # -> VideoFeatures
        pass

    def __call__(self, text: Union[str, Iterable[str]]) -> Any:
        return self.translate(text)


"""Module that contains Sign Languages as classes with rules 
to translate text tokens into sign language videos
"""
from sign_language_translator.languages.sign.mapping_rules import (
    CharacterByCharacterMappingRule,
    DirectMappingRule,
    LambdaMappingRule,
    MappingRule,
)
from sign_language_translator.languages.sign.pakistan_sign_language import (
    PakistanSignLanguage,
)
from sign_language_translator.languages.sign.sign_language import SignLanguage

__all__ = [
    "SignLanguage",
    "PakistanSignLanguage",
    "MappingRule",
    "LambdaMappingRule",
    "CharacterByCharacterMappingRule",
    "DirectMappingRule",
]



from __future__ import annotations

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from ..languages.text.text_language import TextLanguage
    from ..languages.sign.sign_language import SignLanguage

import enum


class Countries(enum.Enum):
    PAKISTAN = "pk"
    # USA = "usa"


class Organizations(enum.Enum):
    HFAD = "hfad"
    # NICE = "nice"
    # FESF = "fesf"


class SignCollections(enum.Enum):
    PK_HFAD_1 = f"{Countries.PAKISTAN}-{Organizations.HFAD}-1"
    WORDLESS = "wordless"


class TextLanguages(enum.Enum):
    URDU = "Urdu"
    # ENGLISH = "English"


class SignLanguages(enum.Enum):
    PAKISTAN_SIGN_LANGUAGE = "PakistanSignLanguage"


class VideoFeatures(enum.Enum):
    # Body Landmarks
    # 3d world coordinates (x,y,z,[visibility])
    MEDIAPIPE_POSE_V2_HAND_V1_3D = "mediapipe_pose_v2_hand_v1_3d"

    # Body Mesh Grid

    # Image Segmentation


class ModelCodes(enum.Enum):
    CONCATENATIVE_SYNTHESIS = "concatenative-synthesis"


__all__ = [
    "Countries",
    "Organizations",
    "SignCollections",
    "TextLanguages",
    "SignLanguages",
    "VideoFeatures",
    "ModelCodes",
]



"""file validates and stores properties of data in a single object

Raises:
    ValueError: unhandled type of words.
    ValueError: unhandled type of person_numbers.
    ValueError: unhandled type of camera_angle.
    ValueError: unknown type of signs_collection.
    ValueError: Invalid type of landmarks_type.
    ValueError: directory to Signs dataset does not exit.
"""

from glob import glob
from os.path import abspath, dirname, exists, join
from typing import List, Union


class SignDataAttributes:
    """validates and stores properties of data in a single object"""

    def __init__(
        self,
        words: Union[str, List[str]],
        person_numbers: Union[str, int, List[int]] = "*",
        camera_angles: Union[str, List[str]] = "*",
        signs_collections: Union[str, List[str]] = "*",
        landmarks_type: Union[str, List[str]] = "all_world_landmarks",
        dataset_directory: str = None,
    ):
        """__init__ function. initializes an object containing all attributes of data. '*' means load all available.

        Args:
            words (Union[str, List[str]]): The sign labels to be loaded
            person_numbers (Union[str, int, List[int]], optional): The IDs of people who performed the sign. Defaults to "*".
            camera_angles (Union[str, List[str]], optional): The camera angle with which the sign was recorded. Defaults to "*".
            signs_collections (Union[str, List[str]], optional): The provider of reference clip of the sign. Defaults to "*".
            landmarks_type (Union[str, List[str]], optional): The type of pose vector (body, hand, 2D, 3D). provide from ["Pose_World_Landmarks", "Left_Hand_World_Landmarks", "Right_Hand_World_Landmarks", "Pose_Image_Landmarks", "Left_Hand_Image_Landmarks", "Right_Hand_Image_Landmarks"]. Defaults to "all_world_landmarks".
            dataset_directory (str, optional): where data is contained. Defaults to None.
        """

        self.words = SignDataAttributes.validate_words(words)
        self.person_nos = SignDataAttributes.validate_person_numbers(person_numbers)
        self.camera_angles = SignDataAttributes.validate_camera_angles(camera_angles)
        self.signs_collections = SignDataAttributes.validate_signs_collections(
            signs_collections
        )
        self.landmarks_type = SignDataAttributes.validate_landmarks_type(landmarks_type)
        self.dataset_directory = SignDataAttributes.validate_dataset_directory(
            dataset_directory
        )

        self.landmarks_file_path = join(
            self.dataset_directory, "Landmarks", "mini_mediapipe_landmarks_dataset.zip"
        )

        self.video_file_paths = [
            p
            for word in self.words
            for person_no in self.person_nos
            for camera_angle in self.camera_angles
            for signs_collection in self.signs_collections
            for p in glob(
                join(
                    self.dataset_directory,
                    "Videos",
                    signs_collection,
                    f"person{person_no}",
                    f"{word}_person{person_no}_{camera_angle}.mp4",
                )
            )
        ]

    LANDMARKS_TYPES = [
        # DON'T SHUFFLE
        "Pose_World_Landmarks",
        "Left_Hand_World_Landmarks",
        "Right_Hand_World_Landmarks",
        "Pose_Image_Landmarks",
        "Left_Hand_Image_Landmarks",
        "Right_Hand_Image_Landmarks",
    ]

    WORLD_LANDMARKS_TYPES = LANDMARKS_TYPES[:3]
    IMAGE_LANDMARKS_TYPES = LANDMARKS_TYPES[3:]

    def validate_words(words: Union[str, List[str]]) -> List:
        """validate the provided words and convert to a handled format

        Args:
            words (Union[str, List[str]]): The sign labels to be loaded

        Raises:
            ValueError: unhandled type of words argument

        Returns:
            List: validated sign labels to be loaded
        """

        # warning word not in [...], "unknown words, use from [...]"
        if isinstance(words, str):
            return [words]
        elif isinstance(words, list):
            if "*" in words:
                return ["*"]
            assert all([isinstance(w, str) for w in words]), "use str words"
            return words
        else:
            raise ValueError("unhandled type of words")

    def validate_person_numbers(person_numbers: Union[str, int, List[int]]) -> List:
        """validate the provided person_numbers and convert to a handled format

        Args:
            person_numbers (Union[str, int, List[int]]): The IDs of people who performed the sign.

        Raises:
            ValueError: unhandled type of person_numbers argument

        Returns:
            List: Validated IDs of people who performed the sign.
        """

        # warning person_no not in [...], "unknown person_no, use from [...]"

        if isinstance(person_numbers, str):
            person_numbers = person_numbers.lstrip("person")
            person_numbers = (
                int(person_numbers) if person_numbers != "*" else [person_numbers]
            )

        if isinstance(person_numbers, int):
            return [person_numbers]

        elif isinstance(person_numbers, list):
            if "*" in person_numbers:
                return ["*"]

            assert all(
                [isinstance(n, int) for n in person_numbers]
            ), "use int person_numbers or '*'"

            return person_numbers

        else:
            raise ValueError("unhandled type of person_numbers")

    def validate_camera_angles(camera_angles: Union[str, List[str]]) -> List:
        """validate the provided camera_angles and convert to a handled format

        Args:
            camera_angles (Union[str, List[str]]): The camera angles with which the sign was recorded

        Raises:
            ValueError: unhandled type of camera_angles

        Returns:
            List: Validated camera_angles with which the sign was recorded
        """

        # warning camera_angle not in ["front","left","right","below"], "unknown camera angle, use from ['front', 'left', 'right', 'below']"

        if isinstance(camera_angles, str):
            return [camera_angles]

        elif isinstance(camera_angles, list):
            if "*" in camera_angles:
                return ["*"]

            assert all(
                [isinstance(ca, str) for ca in camera_angles]
            ), "use str camera_angles"

            return camera_angles
        else:
            raise ValueError("unhandled type of camera_angles")

    def validate_signs_collections(signs_collections: Union[str, List[str]]) -> List:
        """validate the provided signs_collections and convert to a handled format

        Args:
            signs_collections (Union[str, List[str]]): The provider of reference clip of the sign

        Raises:
            ValueError: unknown type of signs_collection

        Returns:
            List: Validated signs_collections
        """

        if isinstance(signs_collections, str):
            return [signs_collections]

        if isinstance(signs_collections, list):
            if "*" in signs_collections:
                return ["*"]

            assert set(signs_collections) <= {
                "HFAD_Book1",
                "None",
            }, "signs_collection not right, use from ['*', 'HFAD_Book1', 'None']"

            return signs_collections
        else:
            raise ValueError("unknown type of signs_collection")

    def validate_landmarks_type(landmarks_type: Union[str, List[str]]) -> List:
        """validate the provided landmarks_types and convert to a handled format

        Args:
            landmarks_type (Union[str, List[str]]): The type of pose vector (body, hand, 2D, 3D). provide from ["Pose_World_Landmarks", "Left_Hand_World_Landmarks", "Right_Hand_World_Landmarks", "Pose_Image_Landmarks", "Left_Hand_Image_Landmarks", "Right_Hand_Image_Landmarks"].

        Raises:
            ValueError: invalid landmarks_type, use from ["Pose_World_Landmarks", "Left_Hand_World_Landmarks", "Right_Hand_World_Landmarks", "Pose_Image_Landmarks", "Left_Hand_Image_Landmarks", "Right_Hand_Image_Landmarks"]
            ValueError: Invalid type of landmarks_type. Use list or str

        Returns:
            List: Validated landmarks_types
        """

        if isinstance(landmarks_type, str):
            landmarks_type = [landmarks_type]

        if isinstance(landmarks_type, list):
            landmarks_type_set = set()
            for land_type in landmarks_type:

                if land_type in SignDataAttributes.LANDMARKS_TYPES:
                    landmarks_type_set.add(land_type)

                elif land_type.lower() in ["all_landmarks", "*"]:
                    landmarks_type_set = SignDataAttributes.LANDMARKS_TYPES
                    break

                elif land_type.lower() == "all_world_landmarks":
                    landmarks_type_set |= set(SignDataAttributes.WORLD_LANDMARKS_TYPES)

                elif land_type.lower() == "all_image_landmarks":
                    landmarks_type_set |= set(SignDataAttributes.IMAGE_LANDMARKS_TYPES)

                else:
                    raise ValueError(
                        f"Invalid landmark type. Use from {SignDataAttributes.LANDMARKS_TYPES}"
                    )

            return list(landmarks_type_set)
        else:
            raise ValueError(f"Invalid type of landmarks_type. Use list or str")

    def validate_dataset_directory(dataset_directory: str) -> str:
        """validate the provided dataset_directory and convert to a handled format

        Args:
            dataset_directory (str): directory to Signs dataset

        Raises:
            ValueError: directory to Signs dataset does not exit.

        Returns:
            str: validated dataset_directory
        """

        if dataset_directory is None:
            dataset_directory = abspath(
                join(
                    dirname(dirname(__file__)),
                    "datasets",
                    "Signs_recordings",
                )
            )

        if not exists(dataset_directory):
            raise ValueError(
                f"directory:'{dataset_directory}' to Signs dataset does not exit."
            )

        return dataset_directory



"""# [sign_language_translator](https://github.com/sign-language-translator/sign-language-translator)
This project is an effort to bridge the communication gap between the hearing and the hearing-impaired community using Artificial Intelligence.
The goal is to provide a user friendly API to novel Sign Language Translation solutions that can easily adapt to any regional sign language.

## Usage

```python
import sign_language_translator as slt

# download dataset (by default, dataset is downloaded within the install directory)
# slt.set_dataset_dir("path/to/sign-language-datasets") # optional
# slt.download("id")

# ------------------------------------------

# Load text-to-sign model
# deep_t2s_model = slt.get_model("generative_t2s_base-01") # pytorch
# rule-based model (concatenates clips of each word)
t2s_model = slt.get_model(
    model_code = "ConcatenativeSynthesis", # slt.enums.ModelCodes.CONCATENATIVE_SYNTHESIS.value
    text_language = "English", # or object of any child of slt.languages.text.text_language.TextLanguage class
    sign_language = "PakistanSignLanguage", # or object of any child of slt.languages.sign.sign_language.SignLanguage class
    sign_feature_model = "mediapipe_pose_v2_hand_v1",
)

text = "hello world!"
sign_language_sentence = t2s_model(text)

# moviepy_video_object = sign_language_sentence.video()
# moviepy_video_object.ipython_display()
# moviepy_video_object.write_videofile(f"sentences/{text}.mp4")

# ------------------------------------------

# load sign
video = slt.read_video("video.mp4")
# features = slt.extract_features(video, "mediapipe_pose_v2_hand_v1")

# Load sign-to-text model
deep_s2t_model = slt.get_model("gesture_mp_base-01") # pytorch
features = deep_s2t_model.extract_features(video)

# translate
text = deep_s2t_model(features)
print(text)
```
"""

from sign_language_translator import (
    config,
    data_collection,
    languages,
    models,
    text,
    utils,
    vision,
)
from sign_language_translator.config import enums
from sign_language_translator.config.settings import Settings
from sign_language_translator.config.helpers import set_dataset_dir
from sign_language_translator.languages import get_sign_language, get_text_language
from sign_language_translator.models import get_model

__all__ = [
    "set_dataset_dir",
    "Settings",
    "vision",
    "text",
    "data_collection",
    "models",
    "languages",
    "utils",
    "config",
    "enums",
    # object loaders
    "get_sign_language",
    "get_text_language",
    "get_model"
    # "get_feature_model"
]
